/*
 * $Id: dis_ia64.c 1079 2004-10-22 19:59:34Z troyhebe $
 *
 * This file is part of lcrash, an analysis tool for Linux memory dumps.
 *
 * Created by Silicon Graphics, Inc.
 * Contributions by IBM, and others
 *
 * Copyright (C) 1999 - 2002 Silicon Graphics, Inc. All rights reserved.
 * Copyright (C) 2001, 2002 IBM Deutschland Entwicklung GmbH, IBM Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#include <lcrash.h>
#include <stdarg.h>

/*
 * function declarations
 */
static kaddr_t start_instr_ia64(kaddr_t, int);
static kaddr_t do_dis_ia64(kaddr_t, int, FILE*);
static kaddr_t print_instr_stream_ia64(kaddr_t, int, int, int, FILE*);
static int  print_instr_ia64(kaddr_t, FILE*, int);
static void dump_instr_ia64(kaddr_t, uint64_t, int, FILE *);


#define SLOT_IA64(pc) (((pc & 0xf) < 6) ? 0 : (((pc & 0xf) < 0xc) ? 1 : 2))

/*
 * function definitions
 */

/*
 * start_instr_ia64() - static function to ensure proper padding of instruction
 */
static kaddr_t
start_instr_ia64(kaddr_t pc, int before)
{
	while (before--) {
		if (SLOT_IA64(pc)) {
			pc -= 6;
		} else {
			pc -= 4;
		}
	}
	return(pc);
}

/*
 * do_dis_ia64() - static function to disassemble i instructions 
 */
static kaddr_t
do_dis_ia64(kaddr_t value, int lines, FILE *ofp)
{
	bfd_vma pc;
	int i;

	set_dis_ofp(ofp);

        pc = value;
	for (i = 0; i < lines; i++) {
		dis_printintaddr(pc, &DISINFO, 1);
		pc += print_insn_ia64(pc, &DISINFO);
		DISINFO.fprintf_func(ofp, "\n");
	}
	return((kaddr_t)pc);
}

/*              
 * print_instr_stream_ia64() - static function to disassemble an
 *                             instructions stream
 */
static kaddr_t
print_instr_stream_ia64(kaddr_t v, int bcount, int acount, int flags,
			FILE *ofp)
{
	int count;
	kaddr_t pc = v;

	/* Make sure that output goes to the right place
	 */ 
	set_dis_ofp(ofp);

	/* Make sure that pc is aligned properly
	 */
	if (SLOT_IA64(pc) == 1) {
		pc = ((pc >> 4) << 4) | 0x6;
	} else if (SLOT_IA64(pc) == 2) {
		pc = ((pc >> 4) << 4) | 0xc;
	}
	count = acount + 1;
	if (bcount) {
		pc = start_instr_ia64(pc, bcount);
		count += bcount;	
	}
	pc = do_dis_ia64(pc, count, ofp);
	return(pc);
}

/* 
 * print_instr_ia64()
 */
static int                     
print_instr_ia64(kaddr_t pc, FILE *ofp, int flag)
{               
	return(do_dis_ia64(pc, 1, ofp) - pc);
}

/*
 * dump_instr_ia64() - static function, arch specific instruction dump routine
 */
static void
dump_instr_ia64(kaddr_t addr, uint64_t count, int flags, FILE *ofp)
{
	unsigned char template;
	uint64_t t0, t1, slot0, slot1, slot2;
	kaddr_t bundlep;
	bfd_byte bundle[16];
	
	bundlep = (addr & 0xfffffffffffffff0);
	GET_BLOCK(bundlep, sizeof(bundle), bundle);
	t0 = bfd_getl64(bundle);
	t1 = bfd_getl64(bundle + 8);
	template = (t0 >> 1) & 0xf;
	slot0 = (t0 >>  5) & 0x1ffffffffffLL;
	slot1 = ((t0 >> 46) & 0x3ffff) | ((t1 & 0x7fffff) << 18);
	slot2 = (t1 >> 23) & 0x1ffffffffffLL;

	fprintf(ofp, "\nbundlep  : 0x%"FMTPTR"x\n", bundlep);
	fprintf(ofp, "template : %02x\n", template);
	fprintf(ofp, "   slot0 : %011"FMTPTR"x\n", slot0);
	fprintf(ofp, "   slot1 : %011"FMTPTR"x\n", slot1);
	fprintf(ofp, "   slot2 : %011"FMTPTR"x\n", slot2);
}

/*
 * dis_init_ia64() - init arch specific stuff for disassembling
 */
int
dis_init_ia64(FILE *ofp, int dumparch)
{
	PRINT_INSTR        = print_instr_ia64;
	PRINT_INSTR_STREAM = print_instr_stream_ia64;
	DUMP_INSTR         = dump_instr_ia64;
	USE_OPCODE         = 1;

	DISINFO.fprintf_func   = dis_fprintf;
	DISINFO.stream         = ofp;
	DISINFO.flavour        = bfd_target_elf_flavour;
	DISINFO.arch           = bfd_arch_ia64;
	/* don't know what mach to specify here */
	/* 	DISINFO.mach   = bfd_mach_ia64????; */
	DISINFO.endian         = BFD_ENDIAN_LITTLE;
	DISINFO.read_memory_func 	   = getidmem;
	DISINFO.print_address_func 	   = dis_printaddr;
	DISINFO.symbol_at_address_func = dis_getsym;
	/* XXX check how do we have to set display_endianess here */
        DISINFO.display_endian = BFD_ENDIAN_LITTLE;

	return(0);
}
