#include <glib.h>
#include <string.h>

#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "server.h"
typedef guint32 u32;
typedef guint64 u64;
#include "netconsole.h"
#include "netdumpclient.h"
#include "netdumpelf.h"
#include "configuration.h"

#define MAX_N_RETRIES 50
#define NETDUMP_TIMEOUT_MSEC 100
#define NETDUMP_OUTSTANDING_PACKETS 10
#define NETDUMP_ACKWAIT_MSEC 20

#undef DEBUG

static void netdump_client_remove_timeout (NetdumpClient *client);
static gboolean generate_i386_elf_header (NetdumpClient *client);


static void handshake            (NetdumpClient      *client);
static void handshake_send_hello (NetdumpClient      *client,
				  guint32             request);
static void handshake_packet     (NetdumpClient      *client,
				  struct sockaddr_in *addr,
				  unsigned char      *buffer,
				  int                 buf_len);
static void handshake_timeout    (NetdumpClient      *client,
				  guint32             request);
static void magic_request        (NetdumpClient      *client);
static void magic_packet         (NetdumpClient      *client,
				  struct sockaddr_in *addr,
				  unsigned char      *buffer,
				  int                 buf_len);
static void magic_timeout        (NetdumpClient      *client,
				  guint32             request);
static void nr_pages_request     (NetdumpClient      *client);
static void nr_pages_packet      (NetdumpClient      *client,
				  struct sockaddr_in *addr,
				  unsigned char      *buffer,
				  int                 buf_len);
static void nr_pages_timeout     (NetdumpClient      *client,
				  guint32             request);
static void page_size_request    (NetdumpClient      *client);
static void page_size_packet     (NetdumpClient      *client,
				  struct sockaddr_in *addr,
				  unsigned char      *buffer,
				  int                 buf_len);
static void page_size_timeout    (NetdumpClient      *client,
				  guint32             request);
static void regs_request         (NetdumpClient      *client);
static void regs_packet          (NetdumpClient      *client,
				  struct sockaddr_in *addr,
				  unsigned char      *buffer,
				  int                 buf_len);
static void regs_timeout         (NetdumpClient      *client,
				  guint32             request);
static void memory_request       (NetdumpClient      *client,
				  gboolean            first_time);
static void memory_timeout       (NetdumpClient      *client,
				  guint32             request);
static void memory_packet        (NetdumpClient      *client,
				  struct sockaddr_in *addr,
				  unsigned char      *buffer,
				  int                 buf_len);


void
netdump_remove_arp_cache_entry (NetdumpClient *client)
{
  char *command;
  
  command = g_strdup_printf ("/sbin/arp -d %d.%d.%d.%d",
			     (client->ip >> 24) & 0xff,
			     (client->ip >> 16) & 0xff,
			     (client->ip >> 8) & 0xff,
			     (client->ip) & 0xff);
  system (command);
  g_free (command);
}

void
netdump_client_destroy (NetdumpClient *client)
{
  int i;

#ifdef DEBUG
  g_print (" netdump_client_destroy(%p)\n", client);
#endif

  remove_client (client);
  netdump_client_remove_timeout (client);

  if (client->dump_fd >= 0)
    close (client->dump_fd);
  
  
  for (i = 0; i < client->n_outstanding; i++)
    {
      printf("delete data %d \n", i);
      g_free (client->outstanding[i].data);
    }
   printf("delete outstanding \n");
  g_free (client->outstanding);
  
  printf("delete reg \n");
  g_free (client->regs);
  printf("delete dir \n");
  g_free (client->dump_dir);

  printf("delete client \n");
  g_free (client);
}


guint
netdump_client_hash (gconstpointer key)
{
  const NetdumpClient *client = key;

  return client->ip;
}

gint
netdump_client_compare (gconstpointer a,
			gconstpointer b)
{
  const NetdumpClient *client1 = a;
  const NetdumpClient *client2 = b;

  return (client1->ip == client2->ip);
}

struct TimeoutData {
  NetdumpClient *client;
  guint32 request_nr;
};

static gboolean
netdump_client_timeout (gpointer data)
{
  struct TimeoutData *timeout = data;

  if (timeout->client->timeout)
    timeout->client->timeout (timeout->client, timeout->request_nr);
  
  return FALSE;
}

static void
netdump_client_remove_timeout (NetdumpClient *client)
{
  if (client->timeout_tag)
    g_source_remove (client->timeout_tag);
  client->timeout_tag = 0;
}

void
netdump_client_send_request (NetdumpClient        *client,
			     enum netdump_commands command,
			     guint32               from,
			     guint32               to,
			     gboolean              setup_timeout)
{
  req_t req;
  struct sockaddr_in addr;
  int addr_len = sizeof(addr);
  int len;
  struct TimeoutData *timeout;

#if 0
  g_print ("sending command %d to %d.%d.%d.%d\n",
	   command,
	   (client->ip >> 24) & 0xff,
	   (client->ip >> 16) & 0xff,
	   (client->ip >> 8) & 0xff,
	   (client->ip >> 0) & 0xff);
#endif
  
  addr.sin_family = AF_INET;
  addr.sin_port = htons(config.port);
  addr.sin_addr.s_addr = htonl(client->ip);
  
  memset (&req, 0, sizeof(req));
  memcpy ((char *)&req.magic, client->magic_value, 8);
  req.nr = htonl (++client->outstanding_request);
  req.command = htonl (command);
  req.from = htonl(from);
  req.to = htonl(to);
  
  len = sendto (master_socket,
		&req, sizeof(req), 0,
		(struct sockaddr *)&addr, addr_len);

  if (setup_timeout)
    {
      timeout = g_new (struct TimeoutData, 1);
      timeout->client = client;
      timeout->request_nr = client->outstanding_request;
      client->timeout_tag = 
	g_timeout_add_full (G_PRIORITY_DEFAULT,
			    NETDUMP_TIMEOUT_MSEC,
			    netdump_client_timeout,
			    timeout,
			    g_free);
    }
}


NetdumpClient *
netdump_client_new (guint32     ip,
		    const char *dir,
		    gboolean    just_reboot)
{
  NetdumpClient *client;
  char *command;
  FILE *f;

  client = g_new0 (NetdumpClient, 1);

#ifdef DEBUG
  g_print (" netdump_client_new() ip: %d.%d.%d.%d -> %p\n",
	   (ip >> 24) & 0xff,
	   (ip >> 16) & 0xff,
	   (ip >> 8) & 0xff,
	   (ip) & 0xff,
	   client);
#endif

  client->ip = ip;
  client->dump_dir = g_strdup (dir);
  client->just_reboot = just_reboot;

  client->dump_fd = -1;
  
  client->outstanding_request = 0;

  /* Lame default. should be read from kernel */
  client->page_offset = 0xC0000000;
  
  client->data_offset = 0;
  client->n_timeouts = 0;
  client->n_handshakes = 0;
  handshake (client);

  /* In case we missed the client starting up, if the ip is already
     in the arp cache, make it permanent. */
  command = g_strdup_printf ("/sbin/arp -n  %d.%d.%d.%d",
			     (ip >> 24) & 0xff,
			     (ip >> 16) & 0xff,
			     (ip >> 8) & 0xff,
			     (ip) & 0xff);
  f = popen (command, "r");
  g_free (command);
  
  if (f)
    {
      char line[2048];
      char *p, *hwaddr;

      /* Skip header */
      fgets(line, sizeof (line), f);

      /* Read first line */
      fgets(line, sizeof (line), f);

      p = line;

      /* Skip ip */
      while (*p && !isspace (*p))
	p++;

      /* Skip spaces */
      while (*p && isspace (*p))
	p++;

      if (strncmp (p, "ether", 5) == 0)
	{
	  /* Skip hwtype */
	  while (*p && !isspace (*p))
	    p++;
	  
	  /* Skip spaces */
	  while (*p && isspace (*p))
	    p++;
	  
	  if (*p)
	    {
	      hwaddr = p;
	      
	      /* Skip until end of hwaddr */
	      while (*p && !isspace (*p))
		p++;
	      
	      /* null terminate it. */
	      *p = 0;
	      
	      command = g_strdup_printf ("/sbin/arp -H ether -s  %d.%d.%d.%d %s",
					 (ip >> 24) & 0xff,
					 (ip >> 16) & 0xff,
					 (ip >> 8) & 0xff,
					 (ip) & 0xff,
					 hwaddr);
	      system (command);
	      g_free (command);
	    }
	}
      
      pclose (f);
    }
  
  return client;
}

static gboolean
parse_packet (unsigned char *buffer,
	      int buf_len,
	      reply_t *reply)
{
  if (buf_len < 1 + sizeof (reply_t) ||
      buffer[0] != NETCONSOLE_VERSION)
    return TRUE;
  
  memcpy ((char *)reply, buffer + 1, sizeof(reply_t));

  reply->nr = ntohl (reply->nr);
  reply->code = ntohl (reply->code);
  reply->info = ntohl (reply->info);

#if 0
  g_print ("Parsed reply: nr: %d, code: %d, info: %d\n",
	   reply->nr, reply->code, reply->info);
  g_print ("packet data: '%s'\n", buffer + 1 + sizeof (reply_t));
#endif
  
  return FALSE;
}

/********************************************************************************
 *                       Handshaking                                            *
 ********************************************************************************/

static void
handshake (NetdumpClient *client)
{
  struct TimeoutData *timeout;
  
#ifdef DEBUG
  g_print ("handshake(%p)\n", client);
#endif
  
  netdump_client_send_request (client,
			       COMM_START_NETDUMP_ACK,
			       0, 0, FALSE);

  /* Wait a while for the ACK to arrive at the client */
  timeout = g_new (struct TimeoutData, 1);
  timeout->client = client;
  timeout->request_nr = 0;
  client->timeout_tag = 
    g_timeout_add_full (G_PRIORITY_DEFAULT,
			NETDUMP_ACKWAIT_MSEC,
			netdump_client_timeout,
			timeout,
			g_free);
  
  client->process_packet = NULL;
  printf("Now we Start Hello send request\n");
  client->timeout = handshake_send_hello;
}

static void
handshake_send_hello (NetdumpClient *client,
		      guint32        request)
{
#ifdef DEBUG
  g_print ("handshake_send_hello(%p)\n", client);
#endif
  netdump_client_send_request (client,
			       COMM_HELLO,
			       0, 0, TRUE);

  client->process_packet = handshake_packet;
  client->timeout = handshake_timeout;
}

static void
handshake_packet (NetdumpClient *client,
		  struct sockaddr_in *addr,
		  unsigned char *buffer,
		  int buf_len)
{
  reply_t reply;
#ifdef DEBUG
  g_print ("handshake_packet(%p)\n", client);
#endif

  if (parse_packet (buffer, buf_len, &reply))
    {
      syslog (LOG_WARNING, "Got invalid packet from 0x%08x\n", client->ip);
      return;
    }

  /* If out of order reply, ignore */
  if (reply.nr != client->outstanding_request)
    return;

  /* If strange reply, ignore. The timeout will catch this. */
  printf("We are in hello ack\n");
  if (reply.code != REPLY_HELLO)
    return;

   printf("We are in hello ack right ON\n");
  netdump_client_remove_timeout (client);

  /* Now request the page size */
  client->n_timeouts = 0;
  magic_request (client);
}

static void
handshake_timeout (NetdumpClient *client,
		   guint32        request)
{
  client->n_timeouts++;

#ifdef DEBUG
  g_print ("handshake_timeout(%p)\n", client);
#endif

  if (client->n_timeouts < MAX_N_RETRIES)
    handshake_send_hello (client, 0);
  else
    {
      client->n_handshakes++;

      if (client->n_handshakes < 10)
	{
	  handshake (client);
	}
      else
	{
	  syslog (LOG_WARNING, "Got to many timeouts in handshaking, ignoring client 0x%08x", client->ip);
	  netdump_client_destroy (client);
	}
    }
}

/********************************************************************************
 *                       GET_MAGIC                                              *
 ********************************************************************************/

static void
magic_request (NetdumpClient *client)
{
#ifdef DEBUG
  g_print ("magic_request (%p)\n", client);
#endif
  
  netdump_client_send_request (client,
			       COMM_GET_MAGIC,
			       0, 0, TRUE);

  client->process_packet = magic_packet;
  client->timeout = magic_timeout;
}

static void
magic_packet (NetdumpClient *client,
	      struct sockaddr_in *addr,
	      unsigned char *buffer,
	      int buf_len)
{
  reply_t reply;
  int i;
  
  if (parse_packet (buffer, buf_len, &reply))
    {
      syslog (LOG_WARNING, "Got invalid packet from 0x%08x\n", client->ip);
      return;
    }

  printf("I am in magic packet\n");
  /* If out of order reply, ignore */
  if (reply.nr != client->outstanding_request)
    return;

  /* If strange reply, ignore. The timeout will catch this. */
  if (reply.code != REPLY_MAGIC)
    return;

  memcpy (client->magic_value, buffer + 1 + sizeof (reply_t), 8);
  
  netdump_client_remove_timeout (client);

  /* By now we have everything we need to communicate with the client. */

  if (client->just_reboot)
    {
      execute_script ("netdump-reboot", client->ip, client->dump_dir);
      for (i = 0; i < 30; i++)
	{
	  netdump_client_send_request (client,
				       COMM_REBOOT,
				       0, 0, FALSE);
	  usleep (10);
	}
      
      netdump_remove_arp_cache_entry (client);
      netdump_client_destroy (client);
    }
  else
    {
      /* Now request the page size */
      client->n_timeouts = 0;
      nr_pages_request (client);
    }
}

static void
magic_timeout (NetdumpClient *client,
	       guint32        request)
{
  client->n_timeouts++;

  if (client->n_timeouts < MAX_N_RETRIES)
    magic_request (client);
  else
    {
      syslog (LOG_WARNING, "Got to many timeouts waiting for MAGIC for client 0x%08x, ignoring it", client->ip);
      netdump_client_destroy (client);
    }
}

/********************************************************************************
 *                       NR_PAGES                                               *
 ********************************************************************************/

static void
nr_pages_request (NetdumpClient *client)
{
#ifdef DEBUG
  g_print ("nr_pages_request (%p)\n", client);
#endif
  netdump_client_send_request (client,
			       COMM_GET_NR_PAGES,
			       0, 0, TRUE);

  client->process_packet = nr_pages_packet;
  client->timeout = nr_pages_timeout;
}

static void
nr_pages_packet (NetdumpClient *client,
		 struct sockaddr_in *addr,
		 unsigned char *buffer,
		 int buf_len)
{
  reply_t reply;

  if (parse_packet (buffer, buf_len, &reply))
    {
      syslog (LOG_WARNING, "Got invalid packet from 0x%08x\n", client->ip);
      return;
    }

  /* If out of order reply, ignore */
  if (reply.nr != client->outstanding_request)
    return;

  /* If strange reply, ignore. The timeout will catch this. */
  if (reply.code != REPLY_NR_PAGES)
    return;

  client->nr_pages = reply.info;
  
  netdump_client_remove_timeout (client);

  /* Now request the page size */
  client->n_timeouts = 0;
  page_size_request (client);
}

static void
nr_pages_timeout (NetdumpClient *client,
		  guint32        request)
{
  client->n_timeouts++;

  if (client->n_timeouts < MAX_N_RETRIES)
    nr_pages_request (client);
  else
    {
      syslog (LOG_WARNING, "Got to many timeouts waiting for NR_PAGES for client 0x%08x, ignoring it", client->ip);
      netdump_client_destroy (client);
    }
}

/********************************************************************************
 *                       PAGE_SIZE                                              *
 ********************************************************************************/

static void
page_size_request (NetdumpClient *client)
{
#ifdef DEBUG
  g_print ("page_size_request (%p)\n", client);
#endif
  netdump_client_send_request (client,
			       COMM_GET_PAGE_SIZE,
			       0, 0, TRUE);

  client->process_packet = page_size_packet;
  client->timeout = page_size_timeout;
}

static void
page_size_packet (NetdumpClient *client,
		  struct sockaddr_in *addr,
		  unsigned char *buffer,
		  int buf_len)
{
  reply_t reply;
  
  if (parse_packet (buffer, buf_len, &reply))
    {
      syslog (LOG_WARNING, "Got invalid packet from 0x%08x\n", client->ip);
      return;
    }

  /* If out of order reply, ignore */
  if (reply.nr != client->outstanding_request)
    return;

  /* If strange reply, ignore. The timeout will catch this. */
  if (reply.code != REPLY_PAGE_SIZE)
    return;

  client->page_size = reply.info;

  netdump_client_remove_timeout (client);

  /* Now request memory... */
//  client->n_timeouts = 0;
//  regs_request (client);
  client->n_timeouts = 0;
  memory_request (client, TRUE);
}

static void
page_size_timeout (NetdumpClient *client,
		   guint32        request)
{
  client->n_timeouts++;

  if (client->n_timeouts < MAX_N_RETRIES)
    page_size_request (client);
  else
    {
      syslog (LOG_WARNING, "Got to many timeouts waiting for PAGE_SIZE for client 0x%08x, ignoring it", client->ip);
      netdump_client_destroy (client);
    }
}

/********************************************************************************
 *                       Registers                                              *
 ********************************************************************************/

static void
regs_request (NetdumpClient *client)
{
#ifdef DEBUG
  g_print ("regs_request (%p)\n", client);
#endif
  netdump_client_send_request (client,
			       COMM_GET_REGS,
			       0, 0, TRUE);

  client->process_packet = regs_packet;
  client->timeout = regs_timeout;
}

static void
regs_packet (NetdumpClient *client,
	     struct sockaddr_in *addr,
	     unsigned char *buffer,
	     int buf_len)
{
  reply_t reply;
  int offset;
  
  if (parse_packet (buffer, buf_len, &reply))
    {
      syslog (LOG_WARNING, "Got invalid packet from 0x%08x\n", client->ip);
      return;
    }

  /* If out of order reply, ignore */
  if (reply.nr != client->outstanding_request)
    return;

  /* If strange reply, ignore. The timeout will catch this. */
  if (reply.code != REPLY_REGS)
    return;

  
  client->regs_len = reply.info;

  /* Find end of ascii data, where regs data starts */
  offset = 1 + sizeof (reply_t);
  while (offset < buf_len && buffer[offset] != '\n')
    offset++;

  if (offset < buf_len)
    offset++; /* Skip past the \n */
  
  if (offset < buf_len)
    {
      client->regs_len = buf_len - offset;
      client->regs = g_malloc (client->regs_len);
      memcpy (client->regs, buffer + offset, client->regs_len);
    }

  netdump_client_remove_timeout (client);

  /* Now request memory... */
  client->n_timeouts = 0;
  memory_request (client, TRUE);
}

static void
regs_timeout (NetdumpClient *client,
	      guint32        request)
{
  client->n_timeouts++;

  if (client->n_timeouts < MAX_N_RETRIES)
    regs_request (client);
  else
    {
      syslog (LOG_WARNING, "Got to many timeouts waiting for REGS for client 0x%08x, ignoring it", client->ip);
      netdump_client_destroy (client);
    }
}


/********************************************************************************
 *                       Dump memory                                            *
 ********************************************************************************/

static void
get_page (NetdumpClient *client, OutstandingPage *outstanding, guint32 page_nr)
{
  outstanding->page_nr = page_nr;
  netdump_client_send_request (client,
			       COMM_SEND_MEM,
			       outstanding->page_nr, 0, TRUE);
  outstanding->request_nr = client->outstanding_request;
  outstanding->timeout_tag = client->timeout_tag;
}


static void
get_page_ack (NetdumpClient *client, guint32 page_nr)
{
  netdump_client_send_request (client,
			       COMM_START_DUMP_WRITE_ACK,
			       page_nr, 0, TRUE);
}




static void
memory_request (NetdumpClient *client, gboolean first_time)
{
  int i;
  gboolean has_active;
  OutstandingPage *outstanding;
  gboolean tried_once = FALSE;
  
  printf("ok Give me memory %d %d\n",  client->nr_pages,  client->current_page);
  if (first_time)
    {
      char *file;
      struct statfs buf;

    retry:
      
      printf("I LOVE THIS\n");
      file = g_strconcat (client->dump_dir, "/vmcore-incomplete", NULL);
      printf("I LOVE THIS %s\n", file);	
      
      client->dump_fd = open (file, O_WRONLY | O_TRUNC | O_CREAT, S_IRUSR | S_IWUSR);
      
      if (client->dump_fd < 0)
	{
	  syslog (LOG_ERR, "unable to open crash dump file");
          printf("memory request delete client\n");
	  netdump_client_destroy (client);
	  g_free (file);
	  return;
	}
      
      if (fstatfs(client->dump_fd, &buf) < 0)
	syslog (LOG_WARNING, "unable to statfs on the crash dump file");
      else
	{
	  off_t memdump_size;
	  int res;

	  memdump_size = ((off_t)client->nr_pages * (off_t)client->page_size + buf.f_bsize - 1) / buf.f_bsize;
	  if (memdump_size > buf.f_bavail)
	    {
	      if (!tried_once)
		{
		  res = execute_script ("netdump-nospace", client->ip, client->dump_dir);
		  if (res > 0)
		    {
		      close (client->dump_fd);
		      client->dump_fd = -1;
		      unlink (file);
		      g_free (file);
		      tried_once = TRUE;
		      goto retry;
		    }
		}

	      if (tried_once)
		syslog (LOG_ERR, "No space for dump image, even after calling netdump-nospace");
	      else
		syslog (LOG_ERR, "No space for dump image");
	      
	      netdump_client_destroy (client);
	      g_free (file);
	      return;
	    }
	}

      g_free (file);

//      if (generate_i386_elf_header (client))
//	{
//	  syslog (LOG_WARNING, "unable to dump elf header");
//	  netdump_client_destroy (client);
//	  return;
//	}

      client->current_page = 0;
      client->n_outstanding = NETDUMP_OUTSTANDING_PACKETS;
      client->outstanding = g_new (OutstandingPage, client->n_outstanding);
      for (i = 0; i < client->n_outstanding; i++)
	{
	  client->outstanding[i].request_nr = 0; /* Unused */
	  client->outstanding[i].timeout_tag = 0; /* Unused */
	  client->outstanding[i].data = g_malloc (client->page_size);
	}
    }
  
  
  client->process_packet = memory_packet;
  client->timeout = memory_timeout;

  has_active = FALSE;
  //for (i = 0; i < client->n_outstanding; i++)
  for (i = 0; i < 1; i++)
    {
      if (client->outstanding[i].request_nr != 0)
	has_active = TRUE;
      else if (client->current_page >= 0)
//      else if (client->current_page != client->nr_pages)
	{
	  outstanding = &client->outstanding[i];

	  outstanding->n_timeouts = 0;
	  outstanding->sections_recieved = 0;
	  get_page (client, outstanding, client->current_page);
	  
	  has_active = TRUE;
	}
    }  
}

static OutstandingPage *
find_outstanding (NetdumpClient *client,
		  guint32        request)
{
  OutstandingPage *outstanding = NULL;
  int i;

  for (i = 0; i < client->n_outstanding; i++)
    if (client->outstanding[i].request_nr == request)
      {
	outstanding = &client->outstanding[i];
	break;
      }

  return outstanding;
}





static void
memory_timeout_ack (NetdumpClient *client,
		guint32        request)
{

  client->n_timeouts++;

  if (client->n_timeouts < 100){
    get_page_ack (client, client->current_page);	
  }
  else
    {
      
      printf("remove client %p timeout ack\n", client);
      syslog (LOG_WARNING, "Got to many timeouts waiting for memory page ACK for client 0x%08x, ignoring it", client->ip);
      netdump_client_destroy (client);
    }
}

static void
memory_timeout (NetdumpClient *client,
		guint32        request)
{
  OutstandingPage *outstanding;

  outstanding = find_outstanding (client, request);

  if (outstanding == NULL)
    {
      syslog (LOG_WARNING, "Got bogus (non-expected) GET_PAGE timeout (request nr: %d)\n", request);
      return;
    }
  
  outstanding->n_timeouts++;

  if (outstanding->n_timeouts < MAX_N_RETRIES)
    {
      get_page (client, outstanding, outstanding->page_nr);
    }
  else
    {
      
      printf("remove client %p timeout\n", client);
      syslog (LOG_WARNING, "Got to many timeouts waiting for memory page for client 0x%08x, ignoring it", client->ip);
      netdump_client_destroy (client);
    }
}

static void
memory_packet (NetdumpClient *client,
	       struct sockaddr_in *addr,
	       unsigned char *buffer,
	       int buf_len)
{
  OutstandingPage *outstanding;
  reply_t reply;
  gboolean done;
  uint64_t offset;

  printf("Inside memory_packet %d\n", reply.nr);

  if (parse_packet (buffer, buf_len, &reply))
    {
      syslog (LOG_WARNING, "Got invalid packet from 0x%08x\n", client->ip);
      return;
    }
  if (reply.code == REPLY_START_WRITE_DUMP)
	outstanding = NULL;
  else
	outstanding = find_outstanding (client, reply.nr);
 
  if (reply.code == REPLY_END_NETDUMP)
  {
    	char *old_file;
      	char *new_file;
	int i;
      
     	execute_script ("netdump-reboot", client->ip, client->dump_dir);
      	for (i = 0; i < 30; i++)
	{
		netdump_client_send_request (client,
				       COMM_START_DUMP_WRITE_ACK,
				       0, 0, FALSE);
	  	usleep (10);
	}
      
     	printf("HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH\n"); 

      	old_file = g_strconcat (client->dump_dir, "/vmcore-incomplete", NULL);
      	new_file = g_strconcat (client->dump_dir, "/vmcore", NULL);

      	if (link(old_file, new_file) == 0)
		unlink (old_file);

      	g_free (old_file);
      	g_free (new_file);
      	netdump_remove_arp_cache_entry (client);
      	netdump_client_destroy (client);
	return;
  } 
  /* If out of order reply, ignore */
  if (outstanding == NULL)
  {
    
    if (reply.code == REPLY_START_WRITE_DUMP)
    {
	printf("We got start write dump hanashake \n");
        client->current_page = reply.info;
        netdump_client_remove_timeout (client);
        memory_request (client, FALSE);
    
    }
    	return;
  }	
  /* If strange reply, ignore. The timeout will catch this. */
  if (reply.code != REPLY_RESERVED &&
      (reply.code & 0xff) != REPLY_MEM)
  {
    printf("ERROR IN memory packet\n");
    return;
  }

  done = FALSE;
  
  if ((reply.code & 0xff) == REPLY_MEM)
    {
      /* Top 24 bits of the offset is in the top 24 bits of reply.code */
      offset = reply.info;
      offset |= (((uint64_t) reply.code) & 0xffffff00) << 24;
      printf("net-server: Assemble packets offset:%d pagesise:%d total:%d\n", offset, client->page_size, client->nr_pages);
      outstanding->sections_recieved |= 1 << (offset/1024);

      if ((offset >= 0) && (offset + 1024 <= client->page_size))
	memcpy (outstanding->data + offset,
		buffer + 1 + sizeof (reply_t),
		1024);

      /* Check if we have got all page sections */
      if (outstanding->sections_recieved ==
	  ((1 << client->page_size/1024) - 1))
	{

	  int left, len; 
	  off_t off;

	  off = outstanding->page_nr;

	  if (off == 0)
		off = outstanding->page_nr * client->page_size;

	   printf("net-server: Writing to file now\n");
	  done = TRUE;

	  lseek (client->dump_fd,
		 client->data_offset + off,
		 SEEK_SET);

	  left = client->page_size;
	  while (left > 0)
	    {
	      len = write (client->dump_fd,
			   outstanding->data + (client->page_size - left),
			   left);
	      if (len < 0)
		{
		  syslog (LOG_ERR, "Error writing memory dump\n");
		  netdump_client_destroy (client);
		}

	      left -= len;
	    }

	}
      
    }
  else {
    	done = TRUE; /* Reserved memory. ignore */
        
  }



  if (done)
    {
      outstanding->request_nr = 0; /* unused */
      if (outstanding->timeout_tag)
      	g_source_remove (outstanding->timeout_tag);
      outstanding->timeout_tag = 0;
      netdump_client_remove_timeout (client);

     client->process_packet = memory_packet;
     client->timeout = memory_timeout_ack;
     client->n_timeouts = 0;
     get_page_ack (client,  outstanding->page_nr);



    }
}


/********************************************************************************
 *                       Elf core dumping                                       *
 ********************************************************************************/

#define roundup(x, y)  ((((x)+((y)-1))/(y))*(y))

static size_t
dump_elf32_note (char *buf,
		 Elf32_Word type,
		 char *name,
		 char *desc, int d_len)
{
  Elf32_Nhdr *note;
  size_t len;

  note = (Elf32_Nhdr *)buf;
  note->n_namesz = strlen (name);
  note->n_descsz = d_len;
  note->n_type = type;
  len = sizeof (Elf32_Nhdr);
  
  memcpy (buf + len, name, note->n_namesz);
  len = roundup (len + note->n_namesz, 4);
  
  memcpy (buf + len, desc, note->n_descsz);
  len = roundup (len + note->n_descsz, 4);

  return len;
}


static gboolean
generate_i386_elf_header (NetdumpClient *client)
{
  char *buffer, *ptr;
  Elf32_Ehdr *elf;
  Elf32_Phdr *notes;
  Elf32_Phdr *load;
  size_t offset, len;
  char empty[80];
  struct elf_prpsinfo prpsinfo;
  struct elf_prstatus prstatus;
  
  buffer = g_malloc (sizeof (Elf32_Ehdr) + 2*sizeof (Elf32_Phdr) +
		     4096*2); /* 4096*2 should be enought for the notes + roundup */
  
  offset = 0;
  ptr = buffer;

  /* Elf header */
  elf = (Elf32_Ehdr *)ptr;
  memcpy (elf->e_ident, ELFMAG, SELFMAG);
  elf->e_ident[EI_CLASS] = ELFCLASS32;
  elf->e_ident[EI_DATA]	= ELFDATA2LSB;
  elf->e_ident[EI_VERSION]= EV_CURRENT;
  memset (elf->e_ident+EI_PAD, 0, EI_NIDENT-EI_PAD);

  elf->e_type	 = ET_CORE;
  elf->e_machine  = EM_386;
  elf->e_version  = EV_CURRENT;
  elf->e_entry	  = 0;
  elf->e_phoff	  = sizeof (Elf32_Ehdr);
  elf->e_shoff	  = 0;
  elf->e_flags	  = 0;
  elf->e_ehsize	  = sizeof (Elf32_Ehdr);
  elf->e_phentsize = sizeof (Elf32_Phdr);
  elf->e_phnum	  = 2;
  elf->e_shentsize = 0;
  elf->e_shnum	  = 0;
  elf->e_shstrndx  = 0;

  offset += sizeof (Elf32_Ehdr);
  ptr += sizeof (Elf32_Ehdr);

  /* PT_NOTE */
  
  notes = (Elf32_Phdr *)ptr;
  notes->p_type	  = PT_NOTE;
  notes->p_offset = 0; /* TO BE FILLED IN */
  notes->p_vaddr  = 0;
  notes->p_paddr  = 0;
  notes->p_filesz = 0; /* TO BE FILLED IN */
  notes->p_memsz  = 0;
  notes->p_flags  = 0;
  notes->p_align  = 0;

  offset += sizeof (Elf32_Phdr);
  ptr += sizeof (Elf32_Phdr);

  /* PT_LOAD */
  
  load = (Elf32_Phdr *)ptr;
  load->p_type   = PT_LOAD;
  load->p_offset = 0; /* TO BE FILLED IN */
  load->p_vaddr  = client->page_offset;
  load->p_paddr  = 0;
  load->p_filesz = client->page_size * client->nr_pages;
  load->p_memsz  = load->p_filesz;
  load->p_flags  = PF_R|PF_W|PF_X;
  load->p_align  = client->page_size;

  offset += sizeof (Elf32_Phdr);
  ptr += sizeof (Elf32_Phdr);

  notes->p_offset = offset;
  
  /* NT_PRSTATUS note */
  memset (&prstatus, 0, sizeof (prstatus));
  if (client->regs)
    memcpy (&prstatus.pr_reg, client->regs, client->regs_len);
  len = dump_elf32_note (ptr, NT_PRSTATUS, "CORE",
			 (char *)&prstatus, sizeof (prstatus));
  offset += len;
  ptr += len;
  notes->p_filesz += len;


  /* NT_PRPSINFO note */
  memset(&prpsinfo, 0, sizeof (struct elf_prpsinfo));
  prpsinfo.pr_state	= 0;
  prpsinfo.pr_sname	= 'R';
  prpsinfo.pr_zomb	= 0;
  strcpy(prpsinfo.pr_fname, "vmlinux");
  
  len = dump_elf32_note (ptr, NT_PRPSINFO, "CORE",
			 (char *)&prpsinfo, sizeof (prpsinfo));
  
  offset += len;
  ptr += len;
  notes->p_filesz += len;

  /* NT_TASKSTRUCT note */
  /* TODO: Just dump zeros here right now.
   * I don't even know the size of task_struct, just use 80*/
  memset (empty, 0, 80);
  len = dump_elf32_note (ptr, NT_TASKSTRUCT, "CORE",
			 empty, 80);
  offset += len;
  ptr += len;
  notes->p_filesz += len;

  /* Page align offset */
  offset = roundup (offset, client->page_size);

  load->p_offset = offset;
  client->data_offset = offset;

  while (offset > 0)
    {
      len = write (client->dump_fd,
		   buffer + (client->data_offset - offset),
		   offset);
      if (len < 0)
	{
	  g_free (buffer);
	  return TRUE;
	}
      
      offset -= len;
    }

  g_free (buffer);
  return FALSE;
}
