/*-*-c++-*-
 * clayeredlayout.cpp
 * Copyright (c) 2002 Michael Geselbracht <let@user.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#include "clayeredlayout.h"
#include "clinegroup.h"
#include "cedge.h"

#include <set>
#include <queue>
#include <algorithm>
#include <limits.h>
#include <assert.h>

class NodeXComparator {
public:
	bool operator()(Node x, Node y) {
		return (x->x() < y->x()) || (x->x() > y->x() && x->x() < y->x() + y->width());
	}
};

class NodeYComparator {
public:
	bool operator()(Node x, Node y) {
		return (x->y() < y->y());
	}
};

IntPair::IntPair()
{
}

IntPair::IntPair(int a, int b) : first(a), second(b)
{
}

bool operator<(const IntPair& x, const IntPair& y)
{
	if (x.first == y.first) {
		return x.second < y.second;
	}
	
	return x.first < y.first;
}

bool operator<=(const IntPair& x, const IntPair& y)
{
	if (x.first == y.first) {
		return x.second <= y.second;
	}
	
	return x.first <= y.first;
}

bool operator==(const IntPair& x, const IntPair& y)
{
	if (x.first == y.first) {
		return x.second == y.second;
	}
	
	return x.first == y.first;
}

bool operator>=(const IntPair& x, const IntPair& y)
{
	if (x.first == y.first) {
		return x.second >= y.second;
	}
	
	return x.first >= y.first;
}

bool operator>(const IntPair& x, const IntPair& y)
{
	if (x.first == y.first) {
		return x.second > y.second;
	}
	
	return x.first > y.first;
}

CLayeredLayout::CLayeredLayout()
{
}

CLayeredLayout::CLayeredLayout(Graph G) : CLayout(G)
{
}

CLayeredLayout::~CLayeredLayout()
{
}

void
CLayeredLayout::layout()
{
	if (!checkGraph("CLayeredLayout::layout()")) {
		return;
	}
	
	if (!oGraph->isAcyclic()) {
		makeAcyclic();
	}
	makeLayered();
	layerAssignment();
	crossingReduction();
	setCoordinates();
}

void
CLayeredLayout::layoutObject(Node v)
{
	bool fixed = oGraph->fixedPos();
	
	oGraph->setFixedPos(true);
	
	// search predecessor from v in node table
	Node u = v->prev();
	assert(u != 0);
	const int cols = oNodeTable.size();
	for (int n = 0; n < cols; n++) {
		const int size = oNodeTable[n].size();
		for (int i = 0; i < size; i++) {
			if (oNodeTable[n][i] == u) {
				if (n == cols - 1) {
					oNodeTable.resize(oNodeTable.size() + 1);
				}
				oNodeTable[n + 1].push_back(v);
				v->setX(u->x() + u->width() + oGrid.width());
				if (oNodeTable[n + 1].size() > 1) {
					Node w = oNodeTable[n + 1][oNodeTable[n + 1].size() - 2];
					v->setY(w->y() + w->height() + oGrid.height());
				}
				else {
					v->setY(u->y() + u->height() / 2 - v->height() / 2);
				}
			}
		}
	}
	
	crossingReduction();
	setCoordinates();
	oGraph->setFixedPos(fixed);
}

void
CLayeredLayout::geometricLayout(Node v)
{
	bool debug = oDebug;
	bool sigen = signalsBlocked();
	bool fixed = oGraph->fixedPos();
	
	oDebug = false;
	blockSignals(true);
	
	/* Do not move nodes since some have been moved by the user
	*/
	oGraph->setFixedPos(true);
	
	/* Remove virtual nodes adjacent to v.
	*/
	removeVirtualNodes();
	
	/* Create node table based on the geometric positions of the nodes.
	*/
	geometricLayerAssignment();

	/* Create virtual nodes for v if required.
	*/
	makeLayeredFromTable();

	CLayeredLayout::setXCoordinates();
	CLayeredLayout::setYCoordinates();
	geometricLayerAssignment();	
	crossingReduction();	
	setCoordinates();
	
	oGraph->setFixedPos(fixed);
	oDebug = debug;
	blockSignals(sigen);
}

void
CLayeredLayout::makeLayered()
{
	/* create set of start vertices
	*/
	
	std::deque<Node> P;
	Node v;
	
	assert(oGraph != 0 && oGraph->isAcyclic());
	forall_nodes(v, oGraph->nodes()) {
		if (v->indeg() == 0) {
			P.push_back(v);
		}
	}
	
	int newCosts = setLayerCosts(P);
	int totalCosts;
	
	emit sigDebugStage("CLayeredLayout::makeLayered()");
	
	do {
		std::deque<Node> Q = P;
		totalCosts = newCosts;
		
		while (!Q.empty()) {
			Node v = Q.front(); Q.pop_front();
			Edge e;
			
			int indeg = 0;
			forall_edges(e, v->edges()) {
				if (e->from() == v) { // outgoing edges
					Q.push_back(e->to());
				}
				else {
					indeg++;
				}
			}
			
			if (indeg < 2) {
				continue;
			}
			
			// maximum cost to reach v
			int maxCost = v->value();
			
			CEdgeSet::Iterator it = v->edges()->begin();
			while (it != v->edges()->end()) {
				Edge e = *it;
				
				if (e->to() == v && e->value() > 0) { // incoming edge
					if (e->value() < maxCost) { // insert dummy vertices
						int numNodes = maxCost - e->value();
						insertVirtualNodes(numNodes, e);
						++it;
						oGraph->removeEdge(e);
						
						if (oDebug) {
							simpleLayout();
							emit sigDebugStep(QString::number(numNodes) + " " + tr("nodes inserted"));
						}
						
						continue;
					}
				}
				
				++it;
			}
			
			totalCosts = newCosts;
			newCosts = setLayerCosts(P);
		}
	} while (totalCosts < newCosts);
}

int
CLayeredLayout::setLayerCosts(std::deque<Node> Q)
{
	int total = 0;
	
	Node v;
	{
		forall_nodes(v, oGraph->nodes()) {
			v->setValue(0);
		}
	}
	Edge e;
	{
		forall_edges(e, oGraph->edges()) {
			e->setValue(0);
		}
	}
	
	while (!Q.empty()) {
		v = Q.front(); Q.pop_front();
		int cost;
		
		forall_edges(e, v->edges()) {
			if (e->from() == v) {
				cost = v->value() + 1;
				e->setValue(cost);
				
				if (e->to()->value() == 0) {
					Q.push_back(e->to());
				}
				if (e->to()->value() < cost) {				
					e->to()->setValue(cost);
				}
			}
		}
	}
	
	// sum maximum costs
	{
		forall_nodes(v, oGraph->nodes()) {
			total = total + v->value();
		}
	}
	
	return total;
}

void
CLayeredLayout::layerAssignment()
{
	checkGraph();
	CGraph G = *oGraph;
	
//	emit sigDebugStage("CLayeredLayout::layerAssignment()");
	
	/* Find start vertices.
	*/
	std::queue<Node> Q;
	Node v;
	forall_nodes(v, G.nodes()) {
		if (v->indeg() == 0) {
			Q.push(v);
		}
		v->setValue(0);
	}
	
	oNodeTable.clear();
	while (!Q.empty()) {
		Node v = Q.front(); Q.pop();
		layerAssignment(v);
	}
}

void
CLayeredLayout::layerAssignment(Node v)
{
	layerAssignment_priv(v, 0);
}

void
CLayeredLayout::layerAssignment_priv(Node v, int level)
{
	if (oNodeTable.size() <= (unsigned)level) {
		oNodeTable.resize(level + 1);
	}
	
	if (v->value() != 0) {
		return;
	}
	else {
		v->setValue(1);
	}
	
	oNodeTable[level].push_back(v->shadow());
	
	Edge e;
	forall_edges(e, v->edges()) {
		if (e->from() == v) {
			layerAssignment_priv(e->to(), level + 1);
		}
	}
}

void
CLayeredLayout::crossingReduction()
{	
	emit sigDebugStage("CLayeredLayout::crossingReduction()");
	
	NodeTable best = oNodeTable;
	IntPair nc = numCrossings();
	IntPair onc = nc;
	int iter = 0;
	
	for (int i = 0; i < 24; i++) {
		wmedian(iter++);
		transpose();
		
		nc = numCrossings();
		
		if (nc < onc) {
			best = oNodeTable;
			onc = nc;
		}
	}
	
	oNodeTable = best;
}

void
CLayeredLayout::wmedian(int iter)
{
	NodeTable& m = oNodeTable;
	const int maxRank = oNodeTable.size();
	
	if (iter % 2 == 0) {
		for (int n = 1; n < maxRank; n++) {
			const int maxCol = oNodeTable[n].size();
			
			for (int i = 0; i < maxCol; i++) {
				m[n][i]->setValue(findMedian(m[n][i], m[n - 1]));
			}
			
			sort(m[n]);
			
			if (oDebug) {
				CLayeredLayout::setYCoordinates();
				setEdgeCoordinates();
				emit sigDebugStep(tr("Sorted layer") + " " + QString::number(n));
			}
		}
	}
	else {
		for (int n = maxRank - 2; n >= 0; n--) {
			const int maxCol = oNodeTable[n].size();
			
			for (int i = 0; i < maxCol - 1; i++) {
				m[n][i]->setValue(findMedian(m[n][i], m[i + 1]));
			}
			
			sort(m[n]);
			
			if (oDebug) {
				CLayeredLayout::setYCoordinates();
				setEdgeCoordinates();
				emit sigDebugStep(tr("Sorted layer") + " " + QString::number(n));
			}
		}
	}
}

void
CLayeredLayout::transpose()
{
	const int maxRank = oNodeTable.size();
	bool improved = true;
	int count = 0;
	
	while (improved && count++ < 10) {
		improved = false;
		
		for (int n = 0; n < maxRank; n++) {
			const int maxCol = oNodeTable[n].size();
			
			for (int i = 0; i < maxCol - 1; i++) {
				IntPair c1 = numCrossings(n);
				IntPair c2;
				
				if (oGraph->fixedPos()) {
					if (!oNodeTable[n][i]->virtualNode() || !oNodeTable[n][i + 1]->virtualNode()) {
						continue;
					}
				}
							
				Node tmp = oNodeTable[n][i];
				oNodeTable[n][i] = oNodeTable[n][i+1];
				oNodeTable[n][i+1] = tmp;
				c2 = numCrossings(n);
				
				if (c1 > c2) {
					improved = true;
				}
				else {
					oNodeTable[n][i+1] = oNodeTable[n][i];
					oNodeTable[n][i] = tmp;
				}
			}
			
			if (oDebug) {
				CLayeredLayout::setYCoordinates();
				setEdgeCoordinates();
				emit sigDebugStep(QString("CLayeredLayout::transpose(): Rank ") + QString::number(n) + " sorted.");
			}
		}
	}
}

int
CLayeredLayout::findMedian(Node v, std::vector<Node>& col)
{
	std::vector<int> seq;
	
	Edge e;
	forall_edges(e, v->edges()) {
		Node to = e->adjacent(v);
		const int size = col.size();
		
		for (int i = 0; i < size; i++) {
			if (col[i] == to) {
				seq.push_back(i);
				break;
			}
		}
	}
	
/*
	const int size = seq.size();
	int k = 0;
	for (int i = 0; i < size; i++) {
		k = k + seq[i];
	}
	return (size > 0) ? k / size : 0;
*/
	return (seq.size() > 0) ? seq[seq.size() / 2] : 0;
}

void
CLayeredLayout::sort(std::vector<Node>& col)
{
	const int size = col.size();
	
	for (int n = 0; n < size; n++) {
		for (int i = n; i < size - 1; i++) {
		/*
			if (oGraph->fixedPos()) {
				if (!col[i]->virtualNode() || !col[i + 1]->virtualNode()) {
					continue;
				}
			}
		*/
			
			if (col[i]->value() > col[i + 1]->value()) {
				Node tmp = col[i];
				col[i] = col[i + 1];
				col[i + 1] = tmp;
			}
			else if (col[i]->value() == col[i + 1]->value()) {
				if (col[i]->indeg() % 2 && col[i + 1]->indeg() % 2 - 1) {
					Node tmp = col[i];
					col[i] = col[i + 1];
					col[i + 1] = tmp;
				}
			}
		}
	}
	
	if (oGraph->fixedPos()) {
		const int size = oNodeTable.size();
		for (int j = 0; j < size; j++) {
			const int cols = oNodeTable[j].size();
			for (int n = 0; n < cols; n++) {
				for (int i = n; i < cols; i++) {
					if (!oNodeTable[j][i]->virtualNode()) {
						for (int x = i + 1; x < cols; x++) {
							if (!oNodeTable[j][x]->virtualNode()) {
								if (oNodeTable[j][i]->y() > oNodeTable[j][x]->y()) {
									Node tmp = oNodeTable[j][i];
									oNodeTable[j][i] = oNodeTable[j][x];
									oNodeTable[j][x] = tmp;
									break;
								}
							}
						}
					}
				}
			}
		}
	}
}

void
CLayeredLayout::setCoordinates()
{
	setXCoordinates();
	setYCoordinates();
	setEdgeCoordinates();
}

void
CLayeredLayout::setXCoordinates()
{
	NodeTable& m = oNodeTable;
	const int maxRank = m.size();
	
	int x = 10;
	for (int n = 0; n < maxRank; n++) {
		const int maxCol = m[n].size();

		/* maxX is used to determine the maximum width of a column.
			All virtual nodes within the same layer is assigned this width.
			minX contains the position of the leftmost node in a column.
		*/
		
		int maxX = 10;
		int minX = INT_MAX;
		int i;
		for (i = maxCol - 1; i >= 0; i--) {
			if (!m[n][i]->virtualNode()) {
				if (m[n][i]->width() > maxX) {
					maxX = m[n][i]->width();
				}
				if (m[n][i]->x() < minX) {
					minX = m[n][i]->x();
				}
			}
		}
		
		for (i = 0; i < maxCol; i++) {
			Node v = m[n][i];
			
			if (v->virtualNode()) {
				v->setWidth(maxX);
				v->setX(minX);
			}
			
			if (!oGraph->fixedPos()) {
				v->setX(x);
			}
		}
		
		x += maxX + grid().width();
	}
}

void
CLayeredLayout::setYCoordinates()
{
	NodeTable& m = oNodeTable;
	const int maxRank = oNodeTable.size();
	std::vector<int> fixedNodes;
	
	for (int n = 0; n < maxRank; n++) {
		const int maxCol = m[n].size();
		fixedNodes.clear();
		
		// find fixed nodes in this layer and store their relative positions
		for (int i = 0; i < maxCol; i++) {
			Node v = m[n][i];
			if (!v->virtualNode()) {
				fixedNodes.push_back(i);
			}
		}
		
		int nfn = fixedNodes.size();
		if (oGraph->fixedPos() && nfn > 0) {
			int npn = 0;
			int lastTop = 50;
			int f = 0;
			
			for (int c = 0; c < nfn; c++) {
				f = fixedNodes[c];
				if (f - npn == 0) {
					npn = f + 1;
					lastTop = m[n][f]->y() + m[n][f]->height();
					continue;
				}
				int dist = (m[n][f]->y() - lastTop) / (f - npn);
#ifdef WIN32
				// On WIN32, min is a macro and prefixing std:: doesn't work
				dist = min(dist, grid().height());
#else
				dist = std::min(dist, grid().height());
#endif
				for (int i = npn; i < f; i++) {
					m[n][i]->setY(lastTop);
					lastTop += dist;
				}
				lastTop = m[n][f]->y() + m[n][f]->height();
				npn = f + 1;
			}
			for (int i = npn; i < maxCol; i++) {
				m[n][i]->setY(lastTop);
				lastTop += grid().height();
			}
		} else {
			for (int i = 0; i < maxCol; i++) {
				Node v = m[n][i];
				
				if (i == 0) {
					v->setY(50);
				}
				else {
					v->setY(m[n][i - 1]->y() + m[n][i - 1]->height() + grid().height());
				}
			}
		}
	}
}

void
CLayeredLayout::simpleLayout()
{
	if (!oDebug) {
		return;
	}
	
	layerAssignment();
	setXCoordinates();
	setYCoordinates();
	setEdgeCoordinates();
}

/*
void
CLayeredLayout::setEdgeCoordinates()
{
	Edge e;
	forall_edges(e, graph()->edges()) {
		Node from = e->from();
		Node to = e->to();
		if (from->x() < to->x()) {
			e->setP1(QPoint(from->x() + from->width(), from->y() + from->height()/2));
			e->setP2(QPoint(to->x(), to->y() + to->height()/2));
		}
		else {
			e->setP1(QPoint(from->x(), from->y() + from->height()/2));
			e->setP2(QPoint(to->x() + to->width(), to->y() + to->height()/2));
		}
	}
}
*/

void
CLayeredLayout::setEdgeCoordinates()
{
	oGraph->lines()->clear();
	std::queue<Node> Q;
	Node v;
	forall_nodes(v, graph()->nodes()) {
		if (v->indeg() == 0) {
			Q.push(v);
		}
		v->setValue(0);
	}
	
	while (!Q.empty()) {
		v = Q.front(); Q.pop();
		if (v->value() != 0) {
			continue;
		}
		v->setValue(1);
		
	  	CLineGroup group;
	  	
	  	Edge e;
	  	forall_edges(e, v->edges()) {
	  		if (e->from() != v) {
	  			continue;
	  		}
	  		
  			CLine line;
  			QString label = e->label();
  			bool reversed = e->reversed();
	  		
  			line = calculateLine(e->from(), e->to());
  			line.setLabel(label);
	  		line.setReversed(reversed);
			line.setEdge(e);
	  		if (e->to()->virtualNode()) {
	  			line.setDrawArrow(reversed ? true : false);
	  		}
  			group.insert(line);
	  		
  			if (e->to()->virtualNode()) {
  				Node walker = e->to();
	  			
  				do {
  					if (walker->value() != 0) {
  						walker = walker->next();
  						continue;
  					}
  					
	  				walker->setValue(1);
  					Edge ve;
  					forall_edges(ve, walker->edges()) {
  						if (ve->from() == walker) {
  							line = calculateLine(ve->from(), ve->to());
  							line.setLabel(label);
  							line.setEdge(ve);
	  						line.setReversed(reversed);
  							if (walker->next()->virtualNode()) {
  								line.setDrawArrow(false);
  							}
  							else if (reversed) {
  								line.setDrawArrow(false);
  							}
  							group.insert(line);
  							break;
	  					}
  					}
  					
  					/* Create a line for the virtual node.
  					*/
  					QPoint p1(walker->x(), walker->y() + walker->height() / 2);
					QPoint p2(walker->x() + walker->width(), walker->y() + walker->height() / 2);
  					line.setP1(p1);
  					line.setP2(p2);
  					line.setLabel(label);
  					line.setDrawArrow(false);
  					line.setReversed(reversed);
  					group.insert(line);
  					
  					walker = walker->next();
  				} while (walker->virtualNode());
	  			
	  			Q.push(walker);
  			}
  			else {
  				Q.push(e->to());
	  		}
  		
  			graph()->lines()->insert(group);
  			group.clear();
  		}
	}
}

CLine
CLayeredLayout::calculateLine(Node from, Node to) const
{
	CLine line;
	
	if (from->x() < to->x()) {
		line.setP1(QPoint(from->x() + from->width(), from->y() + from->height()/2));
		line.setP2(QPoint(to->x(), to->y() + to->height()/2));
	}
	else {
		line.setP1(QPoint(from->x(), from->y() + from->height()/2));
		line.setP2(QPoint(to->x() + to->width(), to->y() + to->height()/2));
	}
	
	return line;
}

void
CLayeredLayout::removeNode(Node v)
{
	removeVirtualNodes();
	CLayout::removeNode(v);
}

void
CLayeredLayout::removeEdge(Edge e)
{
	if (!e->from()->virtualNode() && e->to()->virtualNode()) {
		Node walker = e->to();
		while (walker->virtualNode()) {
			Node prev = walker;
			walker = walker->next();
			oGraph->removeNode(prev);
		}
	}
	else {
		CLayout::removeEdge(e);
	}
	
	removeVirtualNodes();
}

void
CLayeredLayout::removeVirtualNodes()
{
	Node v;
	forall_nodes(v, oGraph->nodes()) {
		if (!v->virtualNode()) {
			removeVirtualNodes(v);
		}
	}
}

void
CLayeredLayout::removeVirtualNodes(Node v)
{
	assert(!v->virtualNode());
	
	std::queue<Node> Q;
	Edge e;
	forall_edges(e, v->edges()) {
		if (e->to()->virtualNode()) {
			Node vv = e->to();
			QString label = e->label();
			bool reversed = e->reversed();
			
			do {
				Q.push(vv);
				
				Edge ve;
				forall_edges(ve, vv->edges()) {
					if (ve->from() == vv) {
						vv = ve->to();
						break;
					}
				}
			} while (vv->virtualNode());
			
			e = oGraph->createEdge(v, vv);
			e->setLabel(label);
			e->setReversed(reversed);
		}
		else if (e->from()->virtualNode()) {
			Node vv = e->from();
			QString label = e->label();
			bool reversed = e->reversed();
			
			do {
				Q.push(vv);
				
				Edge ve;
				forall_edges(ve, vv->edges()) {
					if (ve->to() == vv) {
						vv = ve->from();
						break;
					}
				}
			} while (vv->virtualNode());
			
			e = oGraph->createEdge(vv, v);
			e->setLabel(label);
			e->setReversed(reversed);
		}
	}
	
	while (!Q.empty()) {
		oGraph->removeNode(Q.front());
		Q.pop();
	}
}

void
CLayeredLayout::geometricLayerAssignment()
{
	std::vector<Node> nodes;
	Node v;
	
	emit sigDebugStage(tr("Geometric layer assignment"));
	
	forall_nodes(v, oGraph->nodes()) {
		nodes.push_back(v);
	}
	
	/* Sort vector using the x-coordinate.
	*/
	std::sort(nodes.begin(), nodes.end(), NodeXComparator());
	
	std::vector<Node>::iterator it = nodes.begin();
	oNodeTable.clear();
	oNodeTable.resize(1);
	oNodeTable[0].push_back(*it);
	
	Node prev = *it;
	++it;
	
	/* Layer assignment.
	*/
	
	unsigned layer = 0;
	int lx = prev->x() + prev->width();
	for (; it != nodes.end(); ++it) {
		if ((*it)->x() > lx) {
			layer++;
		}
		
		if (oNodeTable.size() < layer + 1) {
			oNodeTable.resize(layer + 1);
		}
		
		oNodeTable[layer].push_back(*it);
		lx = (*it)->x() + (*it)->width();
	}
	
	/* Sort nodes within each layer using the node's y-coordinate.
	*/
	
	const int size = oNodeTable.size();
	for (int n = 0; n < size; n++) {
		std::sort(oNodeTable[n].begin(), oNodeTable[n].end(), NodeYComparator());
	}
	
	if (oDebug) {
		setCoordinates();
		emit sigDebugStep(tr("Finished"));
	}
}

Node
CLayeredLayout::insertVirtualNodes(int number, Edge oe)
{
	/* Store edge attributes.
	*/
	QString label = oe->label();
	bool reversed = oe->reversed();
	
	Node u = oe->from();
	Node v = oe->to();
	std::vector<Node> vnodes(number);
	vnodes[0] = oGraph->createNode();
	vnodes[0]->setVirtualNode(true);
	Edge e = oGraph->createEdge(u, vnodes[0]);
	e->setLabel(label);
	e->setReversed(reversed);
	
	for (int i = 1; i < number; i++) {
		vnodes[i] = oGraph->createNode();
		vnodes[i]->setVirtualNode(true);
		e = oGraph->createEdge(vnodes[i - 1], vnodes[i]);
		e->setLabel(label);
		e->setReversed(reversed);
	}
	
	e = oGraph->createEdge(vnodes[number - 1], v);
	e->setLabel(label);
	e->setReversed(reversed);
	
	return vnodes[0];
}

IntPair
CLayeredLayout::numCrossings(int ll) const
{
	int num = 0; // crossings
	int span = 0; // no. of spanning lines
	const int layers = oNodeTable.size();
	int from = 0;
	int to = layers - 1;
	
	if (layers <= 1) {
		return IntPair(0, 0);
	}
	
	if (ll > -1) {
		from = (ll <= 1) ? 0 : ll - 1;
		to = (ll >= layers - 1) ? layers - 1 : ll + 1;
	}
	
	for (int n = from; n < to; n++) {
		const std::vector<Node>& l1 = oNodeTable[n];
		const std::vector<Node>& l2 = oNodeTable[n + 1];
		const int cols1 = l1.size();
		const int cols2 = l2.size();
		
		/* Store the column of each node in layer l1 and l2.
		*/
		int i;
		for (i = 0; i < cols1; i++) {
			l1[i]->setValue(i);
		}
		for (i = 0; i < cols2; i++) {
			l2[i]->setValue(i);
		}
		
		for (int c = 0; c < cols1; c++) {
			Node v = l1[c];
			int p1 = v->value();
			int p2;
			
			Edge e;
			forall_edges(e, v->edges()) {
				if (e->from() == v) {
					p2 = e->to()->value();
				}
				else {
					p2 = e->from()->value();
				}
				
				if ((p1 > p2) || (p1 == 0 && p2 == 0) || (p1 == cols1 - 1 && p2 == cols2 - 1)) {
					continue;
				}
				
				// edges crossing several columns are more expensive than straight edges
				span += abs(p2 - p1);
				
				for (int i = 0; i < cols1; i++) {
					Node u = l1[i];
					int p3 = u->value();
					int p4;
					
					if (u == v) { // don't compare v with v
						continue;
					}
					
					Edge t;
					forall_edges(t, u->edges()) {
						if (t->from() == u) {
							p4 = t->to()->value();
						}
						else {
							p4 = t->from()->value();
						}
						
						if (p4 == -1) {
							continue;
						}
						
						if (
							(p3 > p1 && p2 > p4)
							|| (p3 < p1 && p2 < p4)
						)
						{
							num++;
						}
					}
				}
			}
		}
		
		for (i = 0; i < cols1; i++) {
			l1[i]->setValue(-1);
		}
	}
	
	return IntPair(num, span);
}

void
CLayeredLayout::makeLayeredFromTable()
{
	Node v;
	forall_nodes(v, oGraph->nodes()) {
		makeLayeredFromTable(v);
	}
}

void
CLayeredLayout::makeLayeredFromTable(Node v)
{
	/* Search v in node table and note the layer of each node.
	*/
	
	const int layers = oNodeTable.size();
	
	for (int n = 0; n < layers; n++) {
		const int cols = oNodeTable[n].size();
		
		for (int i = 0; i < cols; i++) {
			Node u = oNodeTable[n][i];
			
			u->setValue(n);
		}
	}
	
	std::queue<Edge> Q;
	Edge e;
	forall_edges(e, v->edges()) {
		Node u = e->to();
		const int diff = u->value() - v->value();
		
		if (e->from() != v || u->virtualNode() || v->virtualNode()) {
			continue;
		}
		
		if (abs(diff) > 1) {
			int rank;
			const int incr = (diff > 1) ? 1 : -1;
			
			Q.push(e);
			
			Node walker = insertVirtualNodes(abs(diff) - 1, e);
			rank = v->value();
			
			while (walker && walker != u) {
				rank += incr;
				assert(rank >= 1);
				assert(rank < layers - 1);
				assert(walker->virtualNode());
				oNodeTable[rank].push_back(walker);
				walker = walker->next();
			}
		}
	}
	
	while (!Q.empty()) {
		Edge e = Q.front(); Q.pop();
		oGraph->removeEdge(e);
	}
}
