/*-*-c++-*-
 * clayeredlayout.h
 * Copyright (c) 2002 Michael Geselbracht <let@user.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#ifndef CLAYEREDLAYOUT_H
#define CLAYEREDLAYOUT_H

#include "osdep.h"
#include "clayout.h"
#include "cline.h"

class IntPair {
public:
	IntPair();
	IntPair(int, int);
	
	int first;
	int second;
};

/**
  *@author Michael Geselbracht
  */

class CLayeredLayout : public CLayout {
public:
	CLayeredLayout();
	CLayeredLayout(Graph G);
	~CLayeredLayout();
	
	virtual void layout();
	virtual void layoutObject(Node);
	virtual void geometricLayout(Node);
	
	/* Create a properly layered graph,
	*/
	void makeLayered();

	/* Remove a node and also remove its adjacent virtual nodes.
	*/
	void removeNode(Node v);
	
	/* Remove an edge and also remove virtual nodes, if any.
	*/
	void removeEdge(Edge e);
	
protected:
	
	/* Do layer assignment by filling `oNodeTable'.
	*/
	void layerAssignment();
	
	/* Try to reduce the number of crossings by reordering nodes within a layer.
	*/
	void crossingReduction();
	
	/* Calculate the total number of crossings. If `ll' is specified, only the number of crossings
	 * between layers [ll-1,ll+1] are calculated.
	*/
	IntPair numCrossings(int ll=-1) const;
	
	/* Set node coordinates using the node table.
	*/
	void setCoordinates();
	void setXCoordinates();
	void setYCoordinates();
	
	void setEdgeCoordinates();
	virtual CLine calculateLine(Node from, Node to) const;
	
	/* Used for debug mode
	*/
	void simpleLayout();
	
private:
	/* Modify edge values to represent the costs to reach the target node via an edge.
	*/
	int setLayerCosts(std::deque<Node> P);
	
	/* Do layer assignment beginning with node `v'.
	*/
	void layerAssignment(Node v);
	void layerAssignment_priv(Node v, int level);
	
	/* Find the column of the median parent node of `v'.
	*/
	int findMedian(Node v, std::vector<Node>& col);
	
	/* Sort columns of node table in forward (iter is even) or reverse (iter is odd) direction.
	*/
	void wmedian(int iter);
	
	/* Exchange adjacent vertices and check if the number of crossings has been reduced.
	*/
	void transpose();
	
	/* Sort an array of nodes according to their `value' attribute.
	*/
	void sort(std::vector<Node>& col);
	
	/* Remove adjacent virtual nodes.
	*/
	void removeVirtualNodes(Node);
	void removeVirtualNodes();
	
	/* Do layer assignment based on the node coordinates.
	*/
	void geometricLayerAssignment();
	
	/* Insert `n' virtual nodes between nodes `u' and `v' which must be adjacent.
	 * No edges are removed. The first virtual node is returned.
	 * { n > 0 }
	*/
	Node insertVirtualNodes(int n, Edge oe);
	
	/* Insert virtual nodes for v to ensure a properly layered graph.
	 * Requires and modifies the node table.
	*/
	void makeLayeredFromTable(Node v);
	
	void makeLayeredFromTable();
	
protected:
	NodeTable oNodeTable;
};

#endif
