/*-*-c++-*-
 * clayout.cpp
 * Copyright (c) 2002 Michael Geselbracht <let@user.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#include "osdep.h"
#include "clayout.h"
#include "cedge.h"
#include "crashtypes.h"

#include <qvaluelist.h>

#include <assert.h>
#include <limits.h>
#include <vector>
#include <queue>

CLayout::CLayout()
{
	initVars();
}

CLayout::CLayout(Graph G)
{
	initVars();
	setGraph(G);
}

CLayout::~CLayout()
{
	destroy();
}

void
CLayout::initVars()
{
	oGraph = 0;
	oAutoDelete = false;
	oDebug = false;
	oGrid.setWidth(150);
	oGrid.setHeight(30);
}

void
CLayout::destroy()
{
	if (oAutoDelete) {
		delete oGraph;
	}
}

Graph
CLayout::graph() const
{
	return oGraph;
}

void
CLayout::setGraph(Graph G)
{
	destroy();
	oGraph = G;
}

bool
CLayout::autoDelete() const
{
	return oAutoDelete;
}

void
CLayout::setAutoDelete(bool w)
{
	oAutoDelete = w;
}

void
CLayout::layerAssignment(CGraph& G, NodeTable& nt)
{
	std::queue<Node> Q;
	Node v;
	
	/* Find all start vertices and mark all vertices `unvisited'.
	*/
	forall_nodes(v, G.nodes()) {
		v->setValue(0);
		if (v->indeg() == 0) {
			Q.push(v);
		}
	}
	
	/* Begin depth-first-search.
	*/
	while (!Q.empty()) {
		Node v = Q.front(); Q.pop();
		
		nt[0].push_back(v);
		
		Edge e;
		forall_edges(e, v->edges()) {
			if (e->from() == v) {
				layerAssignment_priv(e->to(), nt, 1);
			}
		}
	}
}

void
CLayout::layerAssignment_priv(Node parent, NodeTable& nt, int row)
{
	if (parent->value() != 0) {
		return;
	}
	
	/* Mark node `visited' and add it to the node table.
	*/
	parent->setValue(1);
	nt[row].push_back(parent);
	
	Edge e;
	forall_edges(e, parent->edges()) {
		if (e->from() == parent) {
			layerAssignment_priv(e->to(), nt, row + 1);
		}
	}
}

int
CLayout::layoutMaxX(std::vector<Node>& col)
{
	int x = 0;
	for (int i = col.size() - 1; i >= 0; i--) {
		int w = col[i]->width();
		if (w > x) {
			x = w;
		}
	}
	
	return x;
}

void
CLayout::makeAcyclic()
{
	if (!checkGraph("CLayout::makeAcyclic()")) {
		return;
	}
	
	/* Reverse some edges by using a greedy algorithm.
	*/

	// create temporary graph
	CGraph G = *oGraph;
	
	std::deque<Node> Sl, Sr;
	
	G.debug();
	
	while (!G.nodes()->empty()) {
		Node v;
		
		while ((v = G.sink()) != 0) {
			Sr.push_front(v->shadow());
			G.removeNode(v);
			G.debug();
		}
		while ((v = G.source()) != 0) {
			Sl.push_back(v->shadow());
			G.removeNode(v);
			G.debug();
		}
		if (!G.nodes()->empty()) {
			int od = INT_MIN;
			Node x = 0;
			
			forall_nodes(v, G.nodes()) {
				int d = v->outdeg() - v->indeg();
				if (d > od) {
					od = d;
					x = v;
				}
			}
			
			assert(x != 0);
			Sl.push_back(x->shadow());
			G.removeNode(x);
			G.debug();
		}
	}
	
	std::vector<Node> vec;
	std::deque<Node>::iterator it;
	int count = 0;
	
	// merge (concat) lists in vector
	for (it = Sl.begin(); it != Sl.end(); ++it) {
		vec.push_back(*it);
		(*it)->setValue(count++);
	}
	for (it = Sr.begin(); it != Sr.end(); ++it) {
		vec.push_back(*it);
		(*it)->setValue(count++);
	}
	
	const int len = vec.size();
	for (int i = 1; i < len; i++) {
		std::list<Edge>::iterator it = vec[i]->edges()->begin();
		while (it != vec[i]->edges()->end()) {
			Node ni = vec[i];
			Edge e = *it;
			
			if (e->from() == ni) { // outgoing edges only
				if (e->to()->value() < ni->value()) { // reverse this edge
					QString label = e->label();
					Node to = e->to();
					Edge ne;
					
					++it;
					oGraph->removeEdge(e);
					
					ne = oGraph->createEdge(to, ni);
					assert(ne != 0);
					ne->setReversed(true);
					ne->setLabel(label);
					
					continue;
				}
			}
			
			++it;
		}
	}
}

bool
CLayout::checkGraph(QString caller) const
{
	if (oGraph == 0) {
		PRINT((caller.isEmpty() ? QString("CLayout::checkGraph()") : caller) + ": There is no graph!");
		return false;
	}
	
	return true;
}

QSize
CLayout::grid() const
{
	return oGrid;
}

void
CLayout::setGrid(QSize grid)
{
	oGrid = grid;
}

void
CLayout::setDebugMode(bool w)
{
	oDebug = w;
}

void
CLayout::removeNode(Node v)
{
	if (!checkGraph("CLayout::removeNode()")) {
		return;
	}
	oGraph->removeNode(v);
}

void
CLayout::removeEdge(Edge e)
{
	if (!checkGraph("CLayout::removeEdge()")) {
		return;
	}
	oGraph->removeEdge(e);
}
