/*-*-c++-*-
 * cngraphmanager.cpp
 * Copyright (c) 2002 Michael Geselbracht <let@user.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#include "osdep.h"
#include "cngraphmanager.h"
#include "cngraphobject.h"
#include "clinegroup.h"
#include "cedge.h"
#include "crashtypes.h"
#include "cconfigmanager.h"

#ifndef WIN32
#  include <unistd.h>
#endif
#include <math.h>
#include <assert.h>

#include <iostream>


#include <qtimer.h>
#include <qapplication.h>
#include <qpopupmenu.h>
#include <qcursor.h>
#include <qpen.h>
#include <qpainter.h>

CNGraphManager::CNGraphManager(CConfigManager* man, QWidget* parent, const char* name)
    : QScrollView(parent, name, WRepaintNoErase|WResizeNoErase|WDestructiveClose)
{
    oConfigManager = man;
	enableClipper(true);
	viewport()->setBackgroundColor(white);
	viewport()->setMouseTracking(true);
	initAttributes();
}

CNGraphManager::~CNGraphManager()
{
	// delete all graph objects
	Node v;
	forall_nodes(v, oGraph->nodes()) {
        if (!v->virtualNode()) {
            CNGraphObject* obj = dynamic_cast<CNGraphObject*>(v->widget());
            assert(obj != 0);
            emit sigChildClosed(obj->widget());
        }
	}

    emit sigClosed(this);
	
	delete oGraph;
}

bool
CNGraphManager::eventFilter(QObject* obj, QEvent* e)
{
	if (obj->isA("CNGraphObject")) {
		switch (e->type()) {
			case QEvent::MouseButtonPress:
				processChildButtonPress((CNGraphObject*) obj, (QMouseEvent*) e);
				break;
			case QEvent::MouseButtonRelease:
				processChildButtonRelease((CNGraphObject*) obj, (QMouseEvent*) e);
				break;
			case QEvent::MouseMove:
				processChildMove((CNGraphObject*) obj, (QMouseEvent*) e);
				break;
			case QEvent::Leave:
				if (oMoveState == None || !oButtonPressed) {
					((CNGraphObject*)obj)->unsetCursor();
                    oMoveState = None;
				}
				break;
			default:
				break;
		}
	}
	else if (obj == viewport()) {
     /*
		if (e->type() == QEvent::Leave) {
			if (oMarkedEdge != 0) {
				oMarkedEdge->setMarked(false);
				oMarkedEdge = 0;
				
				updateContents(0, 0, contentsWidth(), contentsHeight());
			}
		}
	*/
	}
	
	return QScrollView::eventFilter(obj, e);
}

void
CNGraphManager::drawContents(QPainter* p, int clipx, int clipy, int clipw, int cliph)
{
	drawEdges(p);
}

void
CNGraphManager::contentsMouseMoveEvent(QMouseEvent* e)
{
	highlightEdge(e->pos());
	
	if (oMoveState == None && oMarkedEdge == 0 && e->state() == LeftButton) { // move viewport
		QPoint p = oOldContentsPos - e->globalPos();
		oOldContentsPos = e->globalPos();
		p *= 2;
		scrollBy(p.x(), p.y());
	}
}

void
CNGraphManager::contentsMousePressEvent(QMouseEvent* e)
{
	if (oMarkedEdge != 0) {
		if (e->button() == RightButton) {
			execEdgePopup(e);
		}
		else if (e->button() == LeftButton) {
			adjustNewObject(oMarkedEdge->from());
		}
		else if (e->button() == MidButton) {
			adjustNewObject(oMarkedEdge->to());
		}
	}
	
	oOldContentsPos = e->globalPos();
}

void
CNGraphManager::execEdgePopup(QMouseEvent* e)
{
	QPopupMenu* pm = new QPopupMenu(this);
	pm->insertItem("Remove edge", 0);
	
	int id = pm->exec(e->globalPos());
	switch (id) {
		case 0:
			removeMarkedEdge();
			break;
	}
		
	delete pm;
}

void
CNGraphManager::removeMarkedEdge()
{
    if (!oMarkedEdge) {
        return;
    }

    Edge e = (*(oMarkedEdge->begin())).edge();
    assert(!e->from()->virtualNode() || !e->to()->virtualNode());

    oLayout->removeEdge(e);
    oLayout->geometricLayout(0); //! no node

    if (oConfigManager->item(CFG_GRP_LAYOUT, CFG_ATT_LAYOUT_AUTOEDGE, DEF_LAYOUT_AUTOEDGE) != 0) {
        layout();
    }

    redraw();
}

void
CNGraphManager::highlightEdge(QPoint pos)
{
	const int px = pos.x();
	const int py = pos.y();
	
	if (px <= 0 || py <= 0) {
		return;
	}
	
	for (CLineSet::const_iterator n = oGraph->lines()->begin(); n != oGraph->lines()->end(); ++n) {
		for (CLineGroup::const_iterator i = (*n).begin(); i != (*n).end(); ++i) {
			double a, b, w1, wd, t;
			const int x1 = (*i).p1().x();
			const int y1 = (*i).p1().y();
			const int x2 = (*i).p2().x();
			const int y2 = (*i).p2().y();
			
			a = x2 - x1;
			b = y2 - y1;
			
			if (fabs(a) > fabs(b)) {
				if (px < x1 && px < x2 || px > x1 && px > x2) {
					continue;
				}
				
				w1 = atan2(a, b);
				a = px - x1;
				t = tan(w1);
				b = a / ((t != 0) ? t : 1.0) + y1;
				wd = fabs(py - b);
			}
			else {
				if (py < y1 && py < y2 || py > y1 && py > y2) {
					continue;
				}
				
				w1 = atan2(b, a);
				b = py - y1;
				t = tan(w1);
				a = b / ((t != 0) ? t : 1.0) + x1;
				wd = fabs(px - a);
			}
			
			if (wd < 15.0) { // hit !
				if (oMarkedEdge != &(*n)) {
					if (oMarkedEdge != 0) {
						oMarkedEdge->setMarked(false);
					}
					oMarkedEdge = (CLineGroup*) &(*n);
					oMarkedEdge->setMarked(true);
					
					updateContents(0, 0, contentsWidth(), contentsHeight());
				}
				
				return;
			}
		}
	}
	
	if (oMarkedEdge != 0) {
		oMarkedEdge->setMarked(false);
		oMarkedEdge = 0;
		
		updateContents(0, 0, contentsWidth(), contentsHeight());
	}
}

void
CNGraphManager::initAttributes()
{
	oGraph = new CGraph;
	oLayout = 0;
	oMarkedEdge = 0;
	oMarkedNode = 0;
	oButtonPressed = false;
	oMoveState = None;
	oMoveTimer = new QTimer(this);
	oMinObjectSize = QSize(206, 150);
	oMaxObjectSize = QSize(800, 1200);
	connect(oMoveTimer, SIGNAL(timeout()), SLOT(slotMoveTimeout()));
	
	viewport()->installEventFilter(this);
}

Node
CNGraphManager::addNode(QWidget* widget, const QString& title)
{
	CNGraphObject* go = new CNGraphObject(title, viewport(), title.latin1());
	go->setWidget(widget);
	go->installEventFilter(this);
	go->setMouseTracking(true);
	go->show();
	addChild(go);
	moveChild(go, widget->x() , widget->y());
	
	Node node = oGraph->createNode();
	node->setWidget(go);
    node->setLabel(title);
	go->setNode(node);
	
	connect(go, SIGNAL(sigClose(CNGraphObject*)), SLOT(slotCloseObject(CNGraphObject*)));
	connect(go, SIGNAL(sigResized(CNGraphObject*)), SLOT(slotChildResized(CNGraphObject*)));
	
	slotClicked(widget);
	
	return node;
}

void
CNGraphManager::removeNode(Node node)
{
	oLayout->removeNode(node);
    CNGraphObject* obj = dynamic_cast<CNGraphObject*>(node->widget());
    emit sigChildClosed(obj->widget());
    delete node->widget();
	
	if (oGraph->nodes()->count() == 0) {
		close();
		return;
	}
	
	oLayout->geometricLayout(0); //! method expects node. Currently, it is just unused
	erase();

    if (oConfigManager->item(CFG_GRP_LAYOUT, CFG_ATT_LAYOUT_AUTOREMOVE, DEF_LAYOUT_AUTOREMOVE) != 0) {
        layout();
    }

	redraw();
}

Edge
CNGraphManager::attach(Node parent, Node child, const QString& label)
{
	Edge e = oGraph->createEdge(parent, child);
	if (!oGraph->isAcyclic()) { // reverse edge
		oGraph->removeEdge(e);
		e = oGraph->createEdge(child, parent);
		assert(e != 0);
		e->setReversed(true);
	}
	assert(e != 0);
	e->setLabel(label);
	
	return e;
}

void
CNGraphManager::detach(Node parent, Node child)
{
	oGraph->removeEdge(parent, child);
}

void
CNGraphManager::processChildButtonPress(CNGraphObject* child, QMouseEvent* e)
{
	if (oButtonPressed) {
		std::cerr << "CNGraphManager::processChildButtonPress(): oButtonPress is already true !" << std::endl;
	}
	
	if (e->button() == LeftButton) {
		if (e->state() != ShiftButton) {
			slotClicked(child->widget());
			oButtonPressed = true;
			oCursPosX = e->pos().x();
			oCursPosY = e->pos().y();
			//child->grabMouse();
		}
		/*
		else if (oMarkedNode != 0) { // testing
			assert(child != 0);
			attach(oMarkedNode->node(), child->node(), QString("Edge ") + QString::number(++oEdgeCounter));
			layout();
			erase();
			repaint();
		}
		*/
	}
}

void
CNGraphManager::processChildButtonRelease(CNGraphObject* child, QMouseEvent* e)
{
	oButtonPressed = false;
//	oMoveState = None;
	child->releaseMouse();
	oMoveTimer->stop();
}

void
CNGraphManager::processChildMove(CNGraphObject* child, QMouseEvent* e)
{
	if (cursor().shape() == WaitCursor) {
		return;
	}
	
	if (!oButtonPressed) {
		QPoint gp = clipper()->mapToGlobal(QPoint(0, 0));
		int tx = childX(child) + gp.x() - contentsX();
		int ty = childY(child) + gp.y() - contentsY();
		gp = QPoint(tx, ty);
		const int gx1 = gp.x();
		const int gy1 = gp.y();
		const int gx2 = gp.x() + child->width();
		const int gy2 = gp.y() + child->height();
		const int gpx = e->globalPos().x();
		const int gpy = e->globalPos().y();
		const int d = 3;
		QCursor curs;
		
		if (gpx <= gx1 + d && gpy <= gy1 + d) {
			curs.setShape(SizeFDiagCursor);
			oMoveState = ResizeLeft | ResizeTop;
		}
		else if (gpx <= gx1 + d && gpy >= gy2 - d) {
			curs.setShape(SizeBDiagCursor);
			oMoveState = ResizeLeft | ResizeBottom;
		}
		else if (gpx >= gx2 - d && gpy <= gy1 + d) {
			curs.setShape(SizeBDiagCursor);
			oMoveState = ResizeRight | ResizeTop;
		}
		else if (gpx >= gx2 - d && gpy >= gy2 - d) {
			curs.setShape(SizeFDiagCursor);
			oMoveState = ResizeRight | ResizeBottom;
		}
		else if (gpx <= gx1 + d) {
			curs.setShape(SizeHorCursor);
			oMoveState = ResizeLeft;
		}
		else if (gpx >= gx2 - d) {
			curs.setShape(SizeHorCursor);
			oMoveState = ResizeRight;
		}
		else if (gpy <= gy1 + d) {
			curs.setShape(SizeVerCursor);
			oMoveState = ResizeTop;
		}
		else if (gpy >= gy2 - d) {
			curs.setShape(SizeVerCursor);
			oMoveState = ResizeBottom;
		}
		else {
			child->setCursor(cursor());
			oMoveState = Move;
			goto pcm_skip;
		}
				
		child->setCursor(curs);
	}
	else { // action
		if (oMoveState == Move) {
		
			/* Move child object.
			*/
			
			int x = contentsX() + clipper()->mapFromGlobal(e->globalPos()).x();
			int y = contentsY() + clipper()->mapFromGlobal(e->globalPos()).y();
			x = x - oCursPosX;
			y = y - oCursPosY;
			if (x < 0) {
				x = 0;
			}
			if (y < 0) {
				y = 0;
			}
			
			moveChild(child, x, y);
			
			child->node()->setX(x);
			child->node()->setY(y);
		}
		else { // resize
			if (child->isMinimized()) {
				return;
			}
			int cx = childX(child);
			int cy = childY(child);
			int cw = child->width();
			int ch = child->height();
			const int px = contentsX() + clipper()->mapFromGlobal(e->globalPos()).x();
			const int py = contentsY() + clipper()->mapFromGlobal(e->globalPos()).y();
			
			if (oMoveState & ResizeRight) {
				cw = px - cx;
			}
			else if (oMoveState & ResizeLeft) {
				cw = cw + cx - px;
				cx = px;
			}
			if (oMoveState & ResizeTop) {
				ch = ch + cy - py;
				cy = py;
			}
			else if (oMoveState & ResizeBottom) {
				ch = py - cy;
			}
			
			if (cw < oMinObjectSize.width()) {
				cw = oMinObjectSize.width();
			}
			else if (cw > oMaxObjectSize.width()) {
				cw = oMaxObjectSize.width();
			}
			if (ch < oMinObjectSize.height()) {
				ch = oMinObjectSize.height();
			}
			else if (ch > oMaxObjectSize.height()) {
				ch = oMaxObjectSize.height();
			}
			
			moveChild(child, cx, cy);
			child->resize(cw, ch);
			
			child->node()->setX(cx);
			child->node()->setY(cy);
			child->node()->setWidth(cw);
			child->node()->setHeight(ch);
		}
		
		/* Move viewport if necessary.
		*/
		
		int xd = 0;
		int yd = 0;
		if (childX(child) + child->width() > contentsX() + visibleWidth()) {
			xd = (childX(child) + child->width()) - (contentsX() + visibleWidth());
		}
		else if (childX(child) <= contentsX()) {
			xd = childX(child) - contentsX();
		}
		if (childY(child) + child->height() > contentsY() + visibleHeight()) {
			yd = (childY(child) + child->height()) - (contentsY() + visibleHeight());
		}
		else if (childY(child) <= contentsY()) {
			yd = childY(child) - contentsY();
		}
		
		if (xd != 0 || yd != 0) {
			oMoveDirection = 0;
				
			if (xd > 0) {
				oMoveDirection |= Right;
			}
			else if (xd < 0) {
				oMoveDirection |= Left;
			}
			if (yd > 0) {
				oMoveDirection |= Down;
			}
			else if (yd < 0) {
				oMoveDirection |= Up;
			}
			
			oMoveTimer->start(50);
		}
		else {
			oMoveTimer->stop();
		}
		
		oLayout->geometricLayout(child->node());
		placeNodes();
		redraw();
	}
	
pcm_skip:

	return;
}

void
CNGraphManager::slotMoveTimeout()
{
	int dx = 0, dy = 0;
	const int diff = 5;

	if (oMoveDirection & Up) {
		dy = -diff;
	}
	else if (oMoveDirection & Down) {
		dy = diff;
	}
	if (oMoveDirection & Left) {
		dx = -diff;
	}
	else if (oMoveDirection & Right) {
		dx = diff;
	}
	
	scrollBy(dx, dy);
}

void
CNGraphManager::placeNodes()
{
	Node v;
	forall_nodes(v, oGraph->nodes()) {
    	if (!v->virtualNode()) {
			moveChild(v->widget(), v->x(), v->y());
		}
	}
}

void
CNGraphManager::drawEdges(QPainter* painter)
{
	for (CLineSet::iterator it = oGraph->lines()->begin(); it != oGraph->lines()->end(); ++it) {
		const CLineGroup& lg = (*it);
		QColor edgeColor;
		if (lg.marked()) {
			edgeColor = red;
		}
		else {
			edgeColor = blue;
		}
		for (CLineGroup::const_iterator git = (*it).begin(); git != (*it).end(); ++git) {
			drawLine(painter, (*git).p1(), (*git).p2(), (*git).drawArrow(), (*git).reversed(), edgeColor);
		}
		
		drawLabel(painter, lg);
	}
}

void
CNGraphManager::drawLine(QPainter* painter, QPoint p1, QPoint p2, bool withArrow, bool reversed, const QColor& color)
{
	painter->setPen(QPen(color, 2));
	painter->setBrush(QBrush::NoBrush);
	painter->drawLine(p1, p2);
	
	if (withArrow) {
		if (reversed) {
			QPoint tmp = p1;
			p1 = p2;
			p2 = tmp;
		}
		QPoint dp = p1 - p2;
		const double a = atan2((double)dp.x(), (double)dp.y());
		const int h = 10;
		const double d = 15.0 / 180.0 * M_PI;
		QPoint ap;
		
		ap.setX(h * sin(a - d));
		ap.setY(h * cos(a - d));
		painter->drawLine(p2, p2 + ap);
		ap.setX(h * sin(a + d));
		ap.setY(h * cos(a + d));
		painter->drawLine(p2, p2 + ap);
	}
}

void
CNGraphManager::drawLabel(QPainter* painter, const CLineGroup& group)
{
	QString label = group[0].label();
/*	
	if (group[0].reversed()) {
		label += "(R)";
	}
*/
	
	/* Determine the median edge/node to place the label
	*/
	
	const int mid = group.size() / 2;
	QPoint p1, p2;
	p1.setX(group[mid].x());
	p1.setY(group[mid].y() + group[mid].height() / 2);
	p2.setX(group[mid].x() + group[mid].width());
	p2.setY(p1.y());
	
	/* Use the bold application font for the label.
	*/
	QFont fnt = font();
	fnt.setBold(true);
	QFontMetrics fm(fnt);
	
	/* Label position.
	*/
	const int fx = p1.x() + (p2 - p1).x() / 2 - fm.width(label) / 2;
	const int fy = p1.y() + (p2 - p1).y() / 2 - fm.height() / 2;
	const int fw = fm.width(label);
	const int fh = fm.height();
	
	/* Finally draw the label.
	*/
	
	painter->save();
	
	painter->setPen(QPen::NoPen);
	painter->setBrush(white);
	painter->drawRect(fx - 1, fy - 1, fw + 2, fh + 2);
	painter->setPen(QPen(black, 1));
	painter->setBrush(QBrush::NoBrush);
	painter->drawText(fx, fy, fw, fh, AlignCenter, label);
	
	painter->restore();
}

void
CNGraphManager::useLayout(CLayout* layout)
{
	delete oLayout;
	
	oLayout = layout;
	oLayout->setGraph(oGraph);
}

void
CNGraphManager::setDebugMode(bool)
{
	// no longer supported
}

void
CNGraphManager::layout()
{
	if (oLayout == 0) {
		PRINT("CNGraphManager::layout(): There is no layout object!");
		return;
	}
	
	oLayout->layout();
	placeNodes();
	redraw();
}

void
CNGraphManager::layoutObject(Node child)
{
	oLayout->layoutObject(child);
	placeNodes();
	redraw();
}

void
CNGraphManager::geometricLayout()
{
	assert(oLayout != 0);
	oLayout->geometricLayout(0); //! no node
	redraw();
}

void
CNGraphManager::slotCloseObject(CNGraphObject* obj)
{
	if (obj == oMarkedNode) {
		oMarkedNode = 0;
	}
	emit sigChildClosed(obj->widget());
	removeNode(obj->node());
}

void
CNGraphManager::erase()
{
	QPainter p(viewport());
	p.eraseRect(viewport()->rect());
}

Node
CNGraphManager::selectedObject() const
{
	return ((oMarkedNode != 0) ? oMarkedNode->node() : 0);
}

void
CNGraphManager::redraw()
{
	erase();
	repaint();
	viewport()->update();
}

void
CNGraphManager::slotClicked(QWidget* w)
{
	if (w == 0) {
		return;
	}
	
	CNGraphObject* obj = dynamic_cast<CNGraphObject*>(w->parentWidget());
	
	if (oMarkedNode != 0) {
		oMarkedNode->setMarked(false);
	}
	
	oMarkedNode = obj;
	oMarkedNode->setMarked(true);
}

void
CNGraphManager::adjustNewObject(Node child)
{
	CNGraphObject* obj = dynamic_cast<CNGraphObject*>(child->widget());
	int x = -1, y = -1;
	
	assert(obj != 0);
	
	// horizontal
	if (childX(obj) + obj->width() > contentsX() + visibleWidth()) {
		x = childX(obj) + obj->width() - visibleWidth() + 10;
	}
	if (childX(obj) < contentsX() || x < contentsX() && x != -1) {
		x = childX(obj) - 10;
	}
	
	// vertical
	if (childY(obj) + obj->height() > contentsY() + visibleHeight()) {
		y = childY(obj) + obj->height() - visibleHeight() + 10;
	}
	if (childY(obj) < contentsY() || y < contentsY() && y != -1) {
		y = childY(obj) - 10;
	}
	
	// new coordinates ?
	if (x != -1 || y != -1) {
		if (x == -1) {
			x = contentsX();
		}
		else if (y == -1) {
			y = contentsY();
		}
		
		scrollBy(x - contentsX(), y - contentsY());
	}
	
	// mark this object
	assert(obj->widget() != 0);
	slotClicked(obj->widget());
}

void
CNGraphManager::setCursor(const QCursor& curs)
{
	Node v;
	forall_nodes(v, oGraph->nodes()) {
		if (v->widget() != 0) {
			v->widget()->setCursor(curs);
		}
	}
	viewport()->setCursor(curs);
}

void
CNGraphManager::unsetCursor()
{
	Node v;
	forall_nodes(v, oGraph->nodes()) {
		if (v->widget() != 0) {
			v->widget()->unsetCursor();
		}
	}
	viewport()->unsetCursor();
}

void
CNGraphManager::slotChildResized(CNGraphObject* obj)
{
	oLayout->geometricLayout(obj->node());
	placeNodes();
	redraw();
}
