/*-*-c++-*-
 * cngraphmanager.h
 * Copyright (c) 2002 Michael Geselbracht <let@user.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#ifndef CNGRAPHMANAGER_H
#define CNGRAPHMANAGER_H

#include <qscrollview.h>

#include "graph.h"
#include "clayout.h"

class CNGraphObject;
class CConfigManager;

/**The new graph manager
  *@author Michael Geselbracht
  */

class CNGraphManager : public QScrollView {
   Q_OBJECT

   enum MoveMode { None=0, Move=1, ResizeLeft=2, ResizeRight=4, ResizeTop=8, ResizeBottom=16 };
   enum Direction { Left=1, Right=2, Up=4, Down=8 };

public:
	CNGraphManager(CConfigManager* man, QWidget* parent=0, const char* name=0);
	~CNGraphManager();
	
	/* Create a new node and assign a widget.
	*/
	Node addNode(QWidget* child, const QString& title);
	
	/* Remove node.
	*/
	void removeNode(Node child);
	
	/* Set layout object. A previous specified layout will be deleted.
	*/
	void useLayout(CLayout*);
	
	/* Layout the whole graph.
	*/
	void layout();
	
	/* Layout a single object
	*/
	void layoutObject(Node child);
	
	void geometricLayout();
	
	/* Add an edge with an label.
	*/
	Edge attach(Node parent, Node child, const QString& label);
	
	/* Remove all edges between parent and child.
	*/
	void detach(Node parent, Node child);
	
	/* Observe mouse events of graph objects and the viewport.
	*/
	bool eventFilter(QObject* obj, QEvent* e);
	
	/* Update graph.
	*/
	void placeNodes();
	void drawEdges(QPainter* painter);
	
	void setDebugMode(bool activate);
	
	void erase();
	
	/* Move viewport to make v visible
	*/
	void adjustNewObject(Node v);
	
	/* Return the currently selected node or null if there is no selected object.
	*/
	Node selectedObject() const;
	
	/* Redraw contents.
	*/
	void redraw();
	
	virtual void setCursor(const QCursor&);
	virtual void unsetCursor();
	
protected:
	virtual void drawContents(QPainter* p, int clipx, int clipy, int clipw, int cliph);
	virtual void contentsMouseMoveEvent(QMouseEvent*);
	virtual void contentsMousePressEvent(QMouseEvent*);
	void initAttributes();
	
	/* Process child events.
	*/
	
	void processChildButtonPress(CNGraphObject* child, QMouseEvent* e);
	void processChildButtonRelease(CNGraphObject* child, QMouseEvent* e);
	void processChildMove(CNGraphObject* child, QMouseEvent* e);
	
	void removeMarkedEdge();
	
	/* Mouse and viewport
	*/
	
	void execEdgePopup(QMouseEvent*);
	
	void drawLine(QPainter* painter, QPoint p1, QPoint p2, bool withArrow, bool reversed, const QColor& color=blue);
	void drawLabel(QPainter* painter, const CLineGroup& group);
	
	void highlightEdge(QPoint pos);

signals:
	/* This graph manager has been closed.
	*/
	void sigClosed(CNGraphManager*);
	
	/* A node has been removed from the graph.
	*/
	void sigChildClosed(QWidget*);
	
public slots:
	/* Mark node corresponding to this widget as selected. `w' is the child of CNGraphObject !
	*/
	void slotClicked(QWidget*);

private slots:
	void slotCloseObject(CNGraphObject*);
	void slotChildResized(CNGraphObject*);
	void slotMoveTimeout();
	
private:
    CConfigManager* oConfigManager;
	Graph oGraph;
	CLayout* oLayout;
	bool oButtonPressed;
	int oMoveState;
	
	/* Relative mouse cursor position when clicking an object.
	*/
	int oCursPosX;
	int oCursPosY;
	
	/* Move direction for viewport.
	*/
	int oMoveDirection;
	
	/* This timer is used to move the viewport.
	*/
	QTimer* oMoveTimer;
	
	/* The currently selected elements.
	*/
	CLineGroup* oMarkedEdge;
	CNGraphObject* oMarkedNode;
	
	/* Mouse coordinate on button press
	*/
	QPoint oOldContentsPos;
	
	/* Minimum/maximum size for child objects
	*/
	QSize oMinObjectSize;
	QSize oMaxObjectSize;
};

#endif
