/*-*-c++-*-
 * cngraphobject.cpp
 * Copyright (c) 2002 Michael Geselbracht <let@user.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#include "cngraphobject.h"
#include "cgraphheader.h"

#include <qcursor.h>

#define FRAMEWIDTH 3
#define HEADER_HEIGHT 15
#define HEADER_COLOR QColor(204, 204, 255)
#define HEADER_COLOR_MARKED QColor(160, 160, 255)

CNGraphObject::CNGraphObject(const QString& label, QWidget* parent, const char* name) : QFrame(parent, name)
{
	initAttributes();
	
	setCaption(label);
	
	oLabel = new CGraphHeader(label, this);
	oLabel->setBackgroundColor(HEADER_COLOR);
	oLabel->move(FRAMEWIDTH, FRAMEWIDTH);
	oLabel->installEventFilter(this);
	oLabel->setMouseTracking(true);
	oLabel->setCursor(ArrowCursor);
	resize(oLabel->width() + 2 * FRAMEWIDTH, HEADER_HEIGHT + 2 * FRAMEWIDTH);
	connect(oLabel, SIGNAL(sigClose()), SLOT(slotClose()));
	connect(oLabel, SIGNAL(sigMinimize()), SLOT(slotMinimize()));
	
	setFrameStyle(Box | Raised);
	setLineWidth(1);
	setMidLineWidth(1);
	setMinimumSize(206, HEADER_HEIGHT + 2 * FRAMEWIDTH);
}

CNGraphObject::~CNGraphObject()
{
}

void
CNGraphObject::resizeEvent(QResizeEvent*)
{
	oLabel->resize(width() - 2 * FRAMEWIDTH, HEADER_HEIGHT);
	oWidget->resize(width() - 2 * FRAMEWIDTH, height() - HEADER_HEIGHT - 2 * FRAMEWIDTH);
}

bool
CNGraphObject::eventFilter(QObject* obj, QEvent* e)
{
	if (
        e->type() == QEvent::MouseButtonPress ||
        e->type() == QEvent::MouseButtonRelease ||
        e->type() == QEvent::MouseMove
    ) {
		QWidget* sw = (QWidget*) obj;
		QMouseEvent* se = (QMouseEvent*) e;
		
		QMouseEvent me(e->type(), sw->mapToParent(se->pos()), se->globalPos(), se->button(), se->state());
		event(&me);
		
		return true;
	}
	
	return QFrame::eventFilter(obj, e);
}

void
CNGraphObject::initAttributes()
{
	oWidget = 0;
	oMarked = false;
	oOneChild = 0;
}

QWidget*
CNGraphObject::widget()
{
	return oWidget;
}

void
CNGraphObject::setWidget(QWidget* child)
{
	QPoint p(FRAMEWIDTH, FRAMEWIDTH + HEADER_HEIGHT);
	
	delete oWidget;
	oWidget = child;
	oWidget->reparent(this, p);
	oWidget->show();
	oWidget->setMouseTracking(true);
	oWidget->installEventFilter(this);
	oWidget->setCursor(arrowCursor);
	resize(oWidget->width() + 2 * FRAMEWIDTH, oWidget->height() + 2 * FRAMEWIDTH + HEADER_HEIGHT);
}

void
CNGraphObject::replaceWidget(QWidget* child, bool keepSize)
{
	QSize size;
	
	if (oWidget != 0) {
		size = oWidget->size();
	}
	
	QPoint p(0, 0);
	delete oWidget;
	oWidget = child;
	child->reparent(this, p, true);
	
	if (keepSize && size.isValid()) {
		oWidget->resize(size);
		oWidget->move(FRAMEWIDTH, FRAMEWIDTH + HEADER_HEIGHT);
	}
	else {
		resize(child->width() + 2 * FRAMEWIDTH, child->height() + HEADER_HEIGHT + 2 * FRAMEWIDTH);
	}
	
	oWidget->move(FRAMEWIDTH, FRAMEWIDTH + HEADER_HEIGHT);
	oWidget->setCursor(arrowCursor);
}

Node
CNGraphObject::node() const
{
	return oNode;
}

void
CNGraphObject::setNode(Node v)
{
	oNode = v;
}

bool
CNGraphObject::marked() const
{
	return oMarked;
}

void
CNGraphObject::setMarked(bool w)
{
	oLabel->setBackgroundColor(w ? HEADER_COLOR_MARKED : HEADER_COLOR);
	oMarked = w;
}

void
CNGraphObject::slotClose()
{
	emit sigClose(this);
}

bool
CNGraphObject::isViewMany() const
{
	return !oLabel->isViewMany();
}

Node
CNGraphObject::oneChild() const
{
	return oOneChild;
}

void
CNGraphObject::setOneChild(Node child)
{
	oOneChild = child;
}

void
CNGraphObject::setCursor(const QCursor& curs)
{
	if (oWidget != 0 && (curs.shape() == WaitCursor || curs.shape() == ArrowCursor)) {
		oWidget->setCursor(curs);
	}
	QFrame::setCursor(curs);
}

void
CNGraphObject::unsetCursor()
{
	if (oWidget != 0) {
		oWidget->unsetCursor();
	}
	QFrame::unsetCursor();
}

void
CNGraphObject::slotMinimize()
{
	toggleMinimize();
}

void
CNGraphObject::toggleMinimize()
{
	if (oOldSize.isValid()) { // expand
		resize(oOldSize);
		oOldSize = QSize(-1, -1);
		oWidget->show();
	}
	else {
		oOldSize = size();
		resize(1, 1);
		oWidget->hide();
	}
	
	emit sigResized(this);
}
