/*-*-c++-*-
 * cnode.h
 * Copyright (c) 2002 Michael Geselbracht <let@user.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#ifndef CNODE_H
#define CNODE_H

#include "common.h"
#include "cedgeset.h"

#include <qstring.h>

class QWidget;

class CGraph;
class CEdge;
class CNode;
typedef CEdge* Edge;
typedef CNode* Node;

struct Node_t {
	int x;			// x-coordinate
	int y;			// y-coordinate
	int width;		// node width
	int height;		// node height
	int value;		// universal attribute
	bool visited;	// the `visited' flag
	bool isvirtual;	// true if this is a virtual node
	Node shadow;	// the corresponding node of a second graph if this graph has been built by a deep copy
	QWidget* widget;	// widget assigned to this node
	CEdgeSet edges;	// edges connecting this node with its neighbours, if any
	QString label;
	
	Node_t() {
		x = y = value = 0;
		width = height = 0;
		visited = false;
		isvirtual = false;
		shadow = 0;
		widget = 0;
	}
};

/**
  *@author Michael Geselbracht
  */

class CNode {
	friend class CGraph;
	
public: 
	CNode();
	CNode(const CNode&);
	~CNode();
	
	void setNode(const CNode&);
	CNode& operator=(const CNode&);
	
	/* The `visited' flag.
	*/
	bool visited() const;
	void setVisited(bool);

	/* All purpose attribute. Basically used by layout algorithms.
	*/
	int value() const;
	void setValue(int v);
	
	/* When creating a deep copy, a shadow node is the correspondig node of the source graph.
	 * Its value is NULL if there is no shadow node.
	*/
	Node shadow();
	void setShadow(Node);
	
	int x() const;
	int y() const;
	void setX(int x);
	void setY(int y);
	
	int width() const;
	int height() const;
	void setWidth(int w);
	void setHeight(int h);
	
	QWidget* widget();
	
	/* Check/specify if this node is a virtual node or not.
	*/
	bool virtualNode() const;
	void setVirtualNode(bool w);
	
	/* Specify a(nother) widget as node. Note that is possible to remove the relationship
	   to a previously assigned widget by passing a NULL pointer. In this case, the last
	   dimensions of that widget are used for further calculations.
	*/
	void setWidget(QWidget*);
	
	/* Node label. If there is a widget assigned to this node this is the same as widget()->caption().
	*/
	QString label() const;
	void setLabel(const QString& lbl);
	
	/* Edges to adjacent nodes;
	*/
	EdgeSet edges();
	
	int degree() const;
	int indeg() const;
	int outdeg() const;
	
	/* Return the next/previous node. If there is more than one
	 * successor/predeccessor an arbitrary one is returned.
	*/
	Node next();
	Node prev();
	
	/* ID for debugging purposes.
	*/
	inline int id() const {
		return oID;
	}
	inline void setID(int id) {
		oID = id;
	}
	
protected:
	void addEdge(Edge e);
	void removeEdge(Edge e);
	
private:
	Node_t oData;
	int oID; // for debugging
};

#endif
