/*-*-c++-*-
 * cpolylinelayout.cpp
 * Copyright (c) 2002 Michael Geselbracht <let@user.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#include "cpolylinelayout.h"
#include "cedge.h"

#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <algorithm>

CPolylineLayout::CPolylineLayout()
{
}

CPolylineLayout::CPolylineLayout(Graph G) : CLayeredLayout(G)
{
}

CPolylineLayout::~CPolylineLayout()
{
}

void
CPolylineLayout::layout()
{
	CLayeredLayout::layout();
	
	setYCoordinates();
	
	setEdgeCoordinates();
}

void
CPolylineLayout::layoutObject(Node v)
{
	bool fixed = oGraph->fixedPos();
	oGraph->setFixedPos(true);
	CLayeredLayout::layoutObject(v);
	setYCoordinates();
	setEdgeCoordinates();
	oGraph->setFixedPos(fixed);
}

void
CPolylineLayout::geometricLayout(Node v)
{
	CLayeredLayout::geometricLayout(v);
	
	bool fixed = oGraph->fixedPos();
	oGraph->setFixedPos(true);
	setYCoordinates();
	setEdgeCoordinates();
	oGraph->setFixedPos(fixed);
}

void
CPolylineLayout::setYCoordinates()
{
	int minDiff = 0;
	int maxDiff;
	int count = 0;
	
	emit sigDebugStage("CPolylineLayout::setYCoordinates()");

	CLayeredLayout::setYCoordinates();
	
	Node v;
	forall_nodes(v, oGraph->nodes()) { // mark `unvisited'
		v->setValue(0);
	}
	
	// begin Rubberband method
	const int layers = oNodeTable.size();
	do {
		maxDiff = minDiff;
		
		for (int n = 0; n < layers; n++) {
			const int cols = oNodeTable[n].size();
			
			for (int i = 0; i < cols; i++) {
				Node v = oNodeTable[n][i];
				int f = force(v);
				int ox = v->y();
				
				if (oGraph->fixedPos() && !v->virtualNode()) {
					continue;
				}
				
				if (f < 0) {
					if (i > 0) {
						Node u = oNodeTable[n][i - 1];
						const int gh = (u->virtualNode() || v->virtualNode()) ? oGrid.height() / 2 : oGrid.height();
						int nx = ox + f; // i would like...
						int ax = u->y() + u->height() + gh;
						
						if (nx > ax) {
							v->setY(nx);
						}
						else {
							v->setY(ax);
						}
					}
					else {
						int nx = ox + f;
						if (nx < 10) {
							nx = 10;
						}
						v->setY(nx);
					}
				}
				else if (f > 0) {
					if (i < cols - 1) {
						Node u = oNodeTable[n][i + 1];
						const int gh = (u->virtualNode() || v->virtualNode()) ? oGrid.height() / 2 : oGrid.height();
						int nx = ox + f;
						int ax = u->y() - gh;
					
						if (nx + v->height() < ax) {
							v->setY(nx);
						}
						else {
							v->setY(ax - v->height());
						}
					}
					else {
						v->setY(ox + f);
					}
				}
				
				v->setValue(1);
				maxDiff = maxDiff + abs(ox - v->y());
			}
		}
		
		if (oDebug) { // do not call simpleLayout() here !
			setXCoordinates();
			setEdgeCoordinates();
			emit sigDebugStep(QString("Y-Coords. ") + QString::number(++count));
		}
	} while (maxDiff > minDiff);
	
	/* Move a region of nodes.
	*/
	
	minDiff = 0;
	do {
		maxDiff = minDiff;
		
		for (int n = 0; n < layers; n++) {
			const int cols = oNodeTable[n].size();
			
			for (int i = 0; i < cols; i++) {
				NodeVector Q;
				
				i = findRegion(i, oNodeTable[n], Q);
				if (!Q.empty()) {
					maxDiff += moveRegion(i, oNodeTable[n], Q);
				}
			}
		}
	} while (maxDiff > minDiff);
	
	// Avoid too much space above the first line
	if (!oGraph->fixedPos()) {
		int min = INT_MAX;
		for (int n = 0; n < layers; n++) {
			if (oNodeTable[n][0]->y() < min) {
				min = oNodeTable[n][0]->y();
			}
		}
		
		if (min > 10) {
			for (int n = 0; n < layers; n++) {
				const int cols = oNodeTable[n].size();
				for (int i = 0; i < cols; i++) {
					Node v = oNodeTable[n][i];
					v->setY(v->y() - (min - 10));
				}
			}
		}
	}
}

int
CPolylineLayout::force(Node v)
{
	int s = 0;
	Edge e;
	int deg = v->degree();
	
	if (deg == 0) {
		deg = 1;
	}
	
	forall_edges(e, v->edges()) {
		Node u = e->adjacent(v);
		
		const int uy = u->y() + u->height() / 2;
		const int vy = v->y() + v->height() / 2;
		
		s = s + (uy - vy);		
		
		if (u->virtualNode() && u->value() != 0) {
			if (v->virtualNode() || v->degree() == 1) {
				s = (uy - vy);
				deg = 1;
				break;
			}
		}
	}
	
	s = s / deg;
	
	return s;
}

int
CPolylineLayout::findRegion(int n, const NodeVector& col, NodeVector& Q)
{
	const int size = col.size();
	int i;
		
	if (oGraph->fixedPos() && !col[n]->virtualNode()) {
		return n + 1;
	}
		
	for (i = n; i < size - 1; i++) {
		if (oGraph->fixedPos() && (!col[i]->virtualNode() || !col[i + 1]->virtualNode())) {
			break;
		}
		if (col[i]->y() + col[i]->height() + oGrid.height() > col[i + 1]->y()) {
			if (force(col[i]) >= force(col[i + 1])) {
				if (i == n) {
					Q.push_back(col[i]);
				}
				Q.push_back(col[i + 1]);
			}
			else {
				break;
			}
		}
		else if (Q.size() > 0) {
			break;
		}
	}
	
	return i + 1;
}

int
CPolylineLayout::moveRegion(int n, const NodeVector& col, NodeVector& Q)
{
	int s = 0;
	int f = 0;
	Node ln;
	
	for (NodeVector::iterator it = Q.begin(); it != Q.end(); ++it) {
		s = s + force(*it);
	}
	assert(Q.size() > 0);
	int size = Q.size();
	s = s / size;
	f = abs(s);
	
	if (s < 0) { // move up
		if (n - Q.size() > 0) { // Q[0] is inner node
			ln = col[n - Q.size() - 1];
#ifdef WIN32
				// On WIN32, min is a macro and prefixing std:: doesn't work
			f = min(f, Q[0]->y() - (ln->y() + ln->height() + oGrid.height()));			
#else
			f = std::min(f, Q[0]->y() - (ln->y() + ln->height() + oGrid.height()));
#endif
		} else { // Q[0] is topmost node -> negative coordinates are not allowed
#ifdef WIN32
			f = min(f, Q[0]->y());
#else
			f = std::min(f, Q[0]->y());
#endif
		}
		for (NodeVector::iterator it = Q.begin(); it != Q.end(); ++it) {
			(*it)->setY((*it)->y() - f);
		}
	}
	else if (s > 0) { // move down
		if (n < (int) col.size() - 1) { // Q[Q.size() - 1] is inner node
			ln = Q[Q.size() - 1];
#ifdef WIN32
			f = min(f, col[n]->y() - (ln->y() + ln->height() + oGrid.height()));
#else
			f = std::min(f, col[n]->y() - (ln->y() + ln->height() + oGrid.height()));
#endif
		}
		for (NodeVector::iterator it = Q.begin(); it != Q.end(); ++it) {
			(*it)->setY((*it)->y() + f);
		}
	}
	
	return f;
}

/*
void
CPolylineLayout::setEdgeCoordinates()
{
	Edge e;
	forall_edges(e, oGraph->edges()) {
		Node u = e->from();
		Node v = e->to();
		QPoint p1;
		QPoint p2;
		
		calculateEdge(u, v, p1, p2);
		e->setP1(p1);
		e->setP2(p2);
		
		if (u->virtualNode()) {
			if (u->x() < v->x()) {
				e->setP1(QPoint(u->x() + u->width(), u->y() + u->height()/2));
			}
			else {
				e->setP1(QPoint(u->x(), u->y() + u->height()/2));
			}
		}
		if (v->virtualNode()) {
			if (u->x() < v->x()) {
				e->setP2(QPoint(v->x(), v->y() + v->height()/2));
			}
			else {
				e->setP2(QPoint(v->x() + v->width(), v->y() + v->height()/2));
			}
		}
	}
}
*/

CLine
CPolylineLayout::calculateLine(Node u, Node v) const
{
	CLine line;
	const int x1 = u->x() + u->width() / 2;
	const int y1 = u->y() + u->height() / 2;
	const int x2 = v->x() + v->width() / 2;
	const int y2 = v->y() + v->height() / 2;
	int xp1 = x1;
	int yp1 = y1;
	int xp2 = x2;
	int yp2 = y2;

	int qdr = findQuadrant(x1, y1, x2, y2);

	if (qdr > 0) {
		int y = int(tanAlpha(x1, y1, x2, y2) * u->width() / 2);
		
		if (y > u->height() / 2) {
			int y = int(tanAlpha(y1, x1, y2, x2) * u->height() / 2);
			
			if (qdr == 2 || qdr == 3) {
				xp1 = x1 + y;
			}
			else {
				xp1 = x1 - y;
			}
			if (qdr == 3 || qdr == 4) {
				yp1 = u->y() + u->height();
			}
			else {
				yp1 = u->y();
			}
		}
		else {
			if (qdr == 1 || qdr == 4) {
				xp1 = u->x();
			}
			else {
				xp1 = u->x() + u->width();
			}
			if (qdr == 1 || qdr == 2) {
				yp1 = y1 - y;
			}
			else {
				yp1 = y1 + y;
			}
		}
	}
	else {
		if (x1 == x2) {
			xp1 = x1;
			yp1 = (y1 > y2) ? u->y() : u->y() + u->height();
		}
		else {
			xp1 = (x1 > x2) ? u->x() : u->x() + u->width();
			yp1 = y1;// / 2;
		}
	}
	
	xp2 = x2;
	yp2 = y2;
	
	// x2,y2
	qdr = findQuadrant(x2, y2, x1, y1);
	if (qdr > 0) {
		int y = int(tanAlpha(x1, y1, x2, y2) * v->width() / 2);
		
		if (y >  v->height() / 2) {
			y = int(tanAlpha(y1, x1, y2, x2) * v->height() / 2);
			if (qdr == 2 || qdr == 3) {
				xp2 = x2 + y;
			}
			else {
				xp2 = x2 - y;
			}
			if (qdr == 3 || qdr == 4) {
				yp2 = v->y() + v->height();
			}
			else {
				yp2 = v->y();
			}
		}
		else {
			if (qdr == 1 || qdr == 4) {
				xp2 = v->x();
			}
			else {
				xp2 = v->x() + v->width();
			}
			if (qdr == 1 || qdr == 2) {
				yp2 = y2 - y;
			}
			else {
				yp2 = y2 + y;
			}
		}
	}
	else {
		if (x1 == x2) {
			xp2 = x2;
			yp2 = (y1 > y2) ? v->y() + v->height() : v->y();
		}
		else {
			xp2 = (x1 > x2) ? v->x() + v->width() : v->x();
			yp2 = y2;
		}
	}
	
	line.setP1(QPoint(xp1, yp1));
	line.setP2(QPoint(xp2, yp2));
	
	if (u->virtualNode()) {
		if (u->x() < v->x()) {
			line.setP1(QPoint(u->x() + u->width(), u->y() + u->height()/2));
		}
		else {
			line.setP1(QPoint(u->x(), u->y() + u->height()/2));
		}
	}
	if (v->virtualNode()) {
		if (u->x() < v->x()) {
			line.setP2(QPoint(v->x(), v->y() + v->height()/2));
		}
		else {
			line.setP2(QPoint(v->x() + v->width(), v->y() + v->height()/2));
		}
	}
	
	return line;
}

int
CPolylineLayout::findQuadrant(int x1, int y1, int x2, int y2) const
{
	int q = 0;
	int dx = x2 - x1;
	int dy = y2 - y1;

	if (dx != 0 && dy != 0) {
		if (dx < 0 && dy < 0)
			q = 1;
		else if (dx > 0 && dy < 0)
			q = 2;
		else if (dx > 0 && dy > 0)
			q = 3;
		else
			q = 4;
	}

	return q;
}

double
CPolylineLayout::tanAlpha(int x1, int y1, int x2, int y2) const
{
	double dx = x2 - x1;
	double dy = y2 - y1;

	return fabs(dy / dx);
}

void
CPolylineLayout::simpleLayout()
{
	if (!oDebug) {
		return;
	}
	
	layerAssignment();
	setXCoordinates();
	
	setEdgeCoordinates();
}
