/*-*-c++-*-
 * cgraph.cpp
 * Copyright (c) 2002 Michael Geselbracht <let@user.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version. See the file COPYING for more
 * information.
 */

#include "osdep.h"
#include "graph.h"
#include "cedge.h"
#include "clinegroup.h"

#include <queue>
#include <map>
#include <iostream>

int CGraph::sID;

CGraph::CGraph()
{
	oNodes = new CNodeSet;
	oEdges = new CEdgeSet;
	oLines = new CLineSet;
	oFixedPos = false;
}

CGraph::CGraph(const CGraph& g)
{
	oNodes = 0;
	oEdges = 0;
	oLines = 0;
	oFixedPos = false;
	
	setGraph(g);
}

CGraph::~CGraph()
{
	destroy();
}

void
CGraph::destroy()
{
	delete oNodes;
	delete oEdges;
	delete oLines;
	
	oNodes = 0;
	oEdges = 0;
	oLines = 0;
}

CGraph&
CGraph::operator=(const CGraph& g)
{
	setGraph(g);
	
	return *this;
}

void
CGraph::setGraph(const CGraph& g)
{
	Graph gp = (Graph) &g;
	NodeSet ns = gp->nodes();
	
	destroy();
	
	oNodes = new CNodeSet;
	oEdges = new CEdgeSet;
	oLines = new CLineSet; // not initialized
	
	std::map<Node, Node> m;
	Node v;
	{
		forall_nodes(v, ns) {
			Node nv = createNode();
			*nv = *v;
			m[v] = nv;
		}
	}
	
	{
		forall_nodes(v, ns) {
			Edge e;
			forall_edges(e, v->edges()) {
				if (e->from() == v) { // outgoing edges only
					createEdge(m[v], m[e->to()]);
				}
			}
		}
	}
}

NodeSet
CGraph::nodes()
{
	return oNodes;
}

EdgeSet
CGraph::edges()
{
	return oEdges;
}

LineSet
CGraph::lines()
{
	return oLines;
}

Node
CGraph::createNode()
{
	Node v = new CNode;
	v->setID(sID++);
	oNodes->add(v);
	
	return v;
}

Edge
CGraph::createEdge(Node from, Node to)
{
	Edge e;
/*
	forall_edges(e, from->edges()) {
		if (e->from() == from && e->to() == to) {
			PRINT("CGraph::createEdge(): Inserting duplicate edge");
			return 0;
		}
	}
*/
	
	e = new CEdge;
	e->setFromTo(from, to);
	oEdges->add(e);
	from->addEdge(e);
	to->addEdge(e);
	
	return e;
}

void
CGraph::removeNode(Node v)
{
	EdgeSet es = v->edges();
	CEdgeSet::Iterator it, nit;
	it = es->begin();
	
	while (it != es->end()) {
		nit = it++;
		removeEdge(*nit);
	}
	
	oNodes->remove(v);
	
	delete v;
}

void
CGraph::removeEdge(Edge e)
{
	oEdges->remove(e);
	e->from()->removeEdge(e);
	e->to()->removeEdge(e);
	
	delete e;
}

void
CGraph::removeEdge(Node parent, Node child)
{
	std::queue<Edge> Q;
	Edge e;
	forall_edges(e, parent->edges()) {
		if (e->to() == child) {
			Q.push(e);
		}
	}
	
	while (!Q.empty()) {
		removeEdge(Q.front());
		Q.pop();
	}
}

bool
CGraph::isAcyclic()
{
	Node v;
	std::queue<Node> Q;
	NodeSet ns = nodes();
	
	forall_nodes(v, ns) {
		if (v->indeg() == 0)
			Q.push(v);
		else
			v->setValue(v->indeg());
	}
	
	int count = 0;
	while (!Q.empty()) {
		Edge e;
		count++;
		v = Q.front(); Q.pop();
		EdgeSet es = v->edges();
		
		forall_edges(e, es) {
			if (e->from() == v) {
				Node to = e->to();
				to->setValue(to->value() - 1);
				if (to->value() == 0) {
					Q.push(to);
				}
			}
		}
	}
	
	return count == nodes()->count();
}

Node
CGraph::sink()
{
	Node v;
	forall_nodes(v, oNodes) {
		if (v->outdeg() == 0) {
			return v;
		}
	}
	
	return 0;
}

Node
CGraph::source()
{
	Node v;
	forall_nodes(v, oNodes) {
		if (v->indeg() == 0) {
			return v;
		}
	}
	
	return 0;
}

void
CGraph::debug()
{
	std::queue<Node> Q;
	Node v;
	
	forall_nodes(v, oNodes) {
		if (v->indeg() == 0) {
			Q.push(v);
		}
		
		v->setValue(0);
	}
	
	while (!Q.empty()) {
		v = Q.front() ; Q.pop();
		if (v->value() != 0) {
			continue;
		}
		v->setValue(1);
		std::cout << v->id() << " : " << v << " (";
		
		Edge e;
		forall_edges(e, v->edges()) {
			if (e->from() == v) {
				Q.push(e->to());
				std::cout << "T " << e->to()->id() << " : " << e->to() << ", ";
			}
			else {
				std::cout << "F " << e->from()->id() << " : " << e->from() << ", ";
			}
		}
		std::cout << ")" << std::endl;
	}
	
	std::cout << std::endl;
}

bool
CGraph::fixedPos() const
{
	return oFixedPos;
}

void
CGraph::setFixedPos(bool w)
{
	oFixedPos = w;
}
